import logging
import os
from copy import copy
import numpy as np
from pathlib import Path
from typing import List, Generator, Callable, ClassVar, Optional, Union, Tuple

from autoconf import cached_property
from autoconf.dictable import to_dict
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.grid.grid_search import make_lists, Sequential
from autofit.non_linear.grid.sensitivity.job import Job, MaskedJobResult
from autofit.non_linear.grid.sensitivity.job import JobResult
from autofit.non_linear.grid.sensitivity.result import SensitivityResult
from autofit.non_linear.parallel import Process
from autofit.text.formatter import write_table


class Sensitivity:
    def __init__(
        self,
        base_model: AbstractPriorModel,
        perturb_model: AbstractPriorModel,
        simulation_instance,
        paths,
        simulate_cls: Callable,
        base_fit_cls: Callable,
        perturb_fit_cls: Callable,
        job_cls: ClassVar = Job,
        batch_range : Tuple[int, int] = None,
        visualizer_cls: Optional[Callable] = None,
        perturb_model_prior_func: Optional[Callable] = None,
        number_of_steps: Union[Tuple[int, ...], int] = 4,
        mask: Optional[List[bool]] = None,
        number_of_cores: int = 2,
        limit_scale: int = 1,
    ):
        """
        Perform sensitivity mapping to evaluate whether a perturbation
        can be detected if it occurs in different parts of an image.

        For a range from 0 to 1 with step_size, for each dimension of the
        perturb_model, a perturbation is created and used in conjunction
        with the instance to create an image.

        For each of these images, a fit is run with just the model and with both
        the model and perturb_model to compare how much better the image
        can be fit if the perturbation is included.

        Parameters
        ----------
        base_model
            A model that fits the instance well
        perturb_model
            A model which provides a perturbations to be applied to the instance
            before creating images
        simulation_instance
            An instance of a model to which perturbations are applied prior to
            images being generated
        simulate_cls
            A class which simulates images from each perturb instance that sensitivity mapping is performed on.
        base_fit_cls
            The class which fits the base model to each simulated dataset of the sensitivity map.
        perturb_fit_cls
            The class which fits the perturb model to each simulated dataset of the sensitivity map.
        batch_range
            The integer range of sensitivity mapping jobs to perform. If None, all jobs are performed. If not None,
            only the jobs with indices within this range are performed. This means, for example, the range can be
            used to distribute jobs to different machines.
        visualizer_cls
            A class which can be used to visualize the results of the sensitivity mapping after each fit is performed,
            therefore providing visualization on the fly.
        number_of_steps
            The number of steps for each dimension of the sensitivity grid. If input as a float the dimensions are
            all that value. If input as a tuple of length the number of dimensions, each tuple value is the number of
            steps in that dimension.
        mask
            A mask to apply to the sensitivity grid, such that all `True` values are not included in the sensitivity
            mapping. This is useful for removing regions of the sensitivity grid that are expected to have no
            sensitivity, for example because they have no signal.
        number_of_cores
            How many cores does this computer have?
        limit_scale
            Scales the priors for each perturbation model.
                A scale of 1 means priors have limits the same size as the grid square.
                A scale of 2 means priors have limits larger than the grid square with
                    width twice a grid square.
                A scale of 0.5 means priors have limits smaller than the grid square
                    with width half a grid square.
        """
        self.logger = logging.getLogger(f"Sensitivity ({paths.name})")

        self.logger.info("Creating")

        self.instance = simulation_instance
        self.model = base_model
        self.perturb_model = perturb_model

        self.paths = paths

        self.simulate_cls = simulate_cls
        self.base_fit_cls = base_fit_cls
        self.perturb_fit_cls = perturb_fit_cls

        self.perturb_model_prior_func = perturb_model_prior_func

        self.job_cls = job_cls
        self.batch_range = batch_range
        self.visualizer_cls = visualizer_cls

        self.number_of_steps = number_of_steps
        self.mask = None

        if mask is not None:
            self.mask = np.asarray(mask)
            if self.shape != self.mask.shape:
                raise ValueError(
                    f"""
                    The mask of the Sensitivity object must have the same shape as the sensitivity grid.
                    
                    For your inputs, the shape of each are as follows:
                    
                    Sensitivity Grid: {self.shape}
                    Mask: {self.mask.shape}
                    """
                )

        self.number_of_cores = number_of_cores

        self.limit_scale = limit_scale

    def run(self) -> SensitivityResult:
        """
        Run fits and comparisons for all perturbations, returning
        a list of results.
        """
        self.logger.info("Running")

        self.paths.save_unique_tag(is_grid_search=True)

        headers = [
            "index",
            *self._headers,
            "log_evidence_increase",
            "log_likelihood_increase",
        ]
        physical_values = list(self._physical_values)

        process_class = Process if self.number_of_cores > 1 else Sequential

        results = []
        jobs = []

        for number in range(len(self._perturb_instances)):
            model = self.model.copy()
            model.perturb = self._perturb_models[number]
            results.append(
                MaskedJobResult(
                    number=number,
                    model=model,
                )
            )

            if not self._should_bypass(number=number):
                jobs.append(self._make_job(number))

        if self.batch_range is not None:
            jobs = jobs[self.batch_range[0]:self.batch_range[1]]

        for result in process_class.run_jobs(
            jobs, number_of_cores=self.number_of_cores
        ):
            if isinstance(result, Exception):
                raise result

            results[result.number] = result

            sensitivity_result = SensitivityResult(
                samples=[result.result.samples_summary for result in results],
                perturb_samples=[
                    result.perturb_result.samples_summary for result in results
                ],
                shape=self.shape,
                path_values=self.path_values,
            )

            if self.visualizer_cls is not None:
                self.visualizer_cls(
                    sensitivity_result=sensitivity_result, paths=self.paths
                )

            os.makedirs(self.paths.output_path, exist_ok=True)

            write_table(
                headers=headers,
                rows=[
                    [
                        result.number,
                        *physical_values[result.number],
                        result.log_evidence_increase,
                        result.log_likelihood_increase,
                    ]
                    for result in results
                ],
                filename=self.results_path,
            )

        # TODO : Had to repeat this code block to get certain unit tests to pass which presumably bypass run_jobs.

        sensitivity_result = SensitivityResult(
            samples=[result.result.samples_summary for result in results],
            perturb_samples=[
                result.perturb_result.samples_summary for result in results
            ],
            shape=self.shape,
            path_values=self.path_values,
        )

        self.paths.save_json("result", to_dict(sensitivity_result))

        return sensitivity_result

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        Returns the shape of the sensitivity grid.

        The shape is the number of steps performed for each dimension of the `perturb_model`. If sensitivity mapping
        is performed in 3D, the shape will therefore be a tuple of length 3.

        The `shape` can vary across dimensions if the `number_of_steps` parameter is input as a tuple.

        Returns
        -------
        The shape of the sensitivity grid.
        """

        if isinstance(self.number_of_steps, tuple):
            return self.number_of_steps

        return tuple(
            self.number_of_steps for _ in range(self.perturb_model.prior_count)
        )

    def shape_index_from_number(self, number: int) -> Tuple[int, ...]:
        """
        Returns the index of the sensitivity grid from a number.

        Parameters
        ----------
        number
            The number of the sensitivity grid.

        Returns
        -------
        The index of the sensitivity grid.
        """
        return np.unravel_index(number, self.shape)

    @property
    def step_size(self) -> Union[float, Tuple]:
        """
        Returns
        -------
        step_size
            The size of a step in any given dimension in hyper space.
        """
        if isinstance(self.number_of_steps, tuple):
            return tuple(
                1 / number_of_steps for number_of_steps in self.number_of_steps
            )
        return 1 / self.number_of_steps

    @property
    def results_path(self) -> Path:
        return self.paths.output_path / "results.csv"

    @cached_property
    def _lists(self) -> List[List[float]]:
        """
        A list of hypercube vectors, used to instantiate
        the perturb_model and create the individual
        perturbations.
        """
        return make_lists(self.perturb_model.prior_count, step_size=self.step_size)

    @cached_property
    def path_values(self):
        paths = [
            self.perturb_model.path_for_prior(prior)
            for prior in self.perturb_model.priors_ordered_by_id
        ]

        return {
            path: list(values) for path, *values in zip(paths, *self._physical_values)
        }

    @cached_property
    def _physical_values(self) -> List[List[float]]:
        """
        Lists of physical values for each grid square
        """
        return [
            [
                prior.value_for(unit_value)
                for prior, unit_value in zip(
                    self.perturb_model.priors_ordered_by_id, unit_values
                )
            ]
            for unit_values in self._lists
        ]

    @cached_property
    def _headers(self) -> Generator[str, None, None]:
        """
        A name for each of the perturb priors
        """
        for path, _ in self.perturb_model.prior_tuples:
            yield path

    @cached_property
    def _labels(self) -> List[str]:
        """
        One label for each perturbation, used to distinguish
        fits for each perturbation by placing them in separate
        directories.
        """
        labels = []
        for list_ in self._lists:
            strings = list()
            for value, prior_tuple in zip(list_, self.perturb_model.prior_tuples):
                path, prior = prior_tuple
                value = prior.value_for(value)
                strings.append(f"{path}_{value}")
            labels.append("_".join(strings))

        return labels

    @cached_property
    def _perturb_instances(self) -> List[ModelInstance]:
        """
        A list of instances each of which defines a perturbation to
        be applied to the image.
        """

        return [
            self.perturb_model.instance_from_unit_vector(list_) for list_ in self._lists
        ]

    @cached_property
    def _perturb_models(self) -> List[AbstractPriorModel]:
        """
        A list of models representing a perturbation at each grid square.

        By default models have priors with limits at the edges of a grid square.
        These limits can be scaled using the limit_scale variable. If the variable
        is 2 then the priors will have width twice the step size.
        """
        if isinstance(self.step_size, tuple):
            step_sizes = self.step_size
        else:
            step_sizes = (self.step_size,) * self.perturb_model.prior_count

        half_steps = [self.limit_scale * step_size / 2 for step_size in step_sizes]

        perturb_models = []
        for list_ in self._lists:
            limits = [
                (
                    prior.value_for(max(0.0, centre - half_step)),
                    prior.value_for(min(1.0, centre + half_step)),
                )
                for centre, prior, half_step in zip(
                    list_,
                    self.perturb_model.priors_ordered_by_id,
                    half_steps,
                )
            ]
            perturb_models.append(self.perturb_model.with_limits(limits))
        return perturb_models

    def _should_bypass(self, number: int) -> bool:
        shape_index = self.shape_index_from_number(number=number)
        return self.mask is not None and np.asarray(self.mask)[shape_index]

    def _make_jobs(self) -> Generator[Job, None, None]:
        for number, _ in enumerate(self._perturb_instances):
            yield self._make_job(number)

    def _make_job(self, number) -> Generator[Job, None, None]:
        """
        Create a list of jobs to be run on separate processes.

        Each job fits a perturb image with the original model
        and a model which includes a perturbation.
        """
        perturb_instance = self._perturb_instances[number]
        perturb_model = self._perturb_models[number]
        label = self._labels[number]

        if self.perturb_model_prior_func is not None:
            perturb_model = self.perturb_model_prior_func(
                perturb_instance=perturb_instance, perturb_model=perturb_model
            )

        simulate_instance = copy(self.instance)
        simulate_instance.perturb = perturb_instance

        paths = self.paths.for_sub_analysis(
            label,
        )

        return self.job_cls(
            simulate_instance=simulate_instance,
            model=self.model,
            perturb_model=perturb_model,
            base_instance=self.instance,
            simulate_cls=self.simulate_cls,
            base_fit_cls=self.base_fit_cls,
            perturb_fit_cls=self.perturb_fit_cls,
            paths=paths,
            number=number,
        )



import numpy as np
import logging
import os
import sys
from typing import Dict, Optional, Tuple

from autofit.database.sqlalchemy_ import sa

from autoconf import conf
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior.vectorized import PriorVectorized
from autofit.non_linear.fitness import Fitness
from autofit.non_linear.paths.null import NullPaths
from autofit.non_linear.search.nest import abstract_nest
from autofit.non_linear.samples.sample import Sample
from autofit.non_linear.samples.nest import SamplesNest


logger = logging.getLogger(__name__)

class Nautilus(abstract_nest.AbstractNest):
    __identifier_fields__ = (
        "n_live",
        "n_update",
        "enlarge_per_dim",
        "n_points_min",
        "split_threshold",
        "n_networks",
        "n_like_new_bound",
        "seed",
        "n_shell",
        "n_eff",
    )

    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        iterations_per_quick_update: Optional[int] = None,
        iterations_per_full_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs
    ):
        """
        A Nautilus non-linear search.

        Nautilus is an optional requirement and must be installed manually via the command `pip install ultranest`.
        It is optional as it has certain dependencies which are generally straight forward to install (e.g. Cython).

        For a full description of Nautilus checkout its Github and documentation webpages:

        https://github.com/johannesulf/nautilus
        https://nautilus-sampler.readthedocs.io/en/stable/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        iterations_per_full_update
            The number of iterations performed between update (e.g. output latest model to hard-disk, visualization).
        number_of_cores
            The number of cores sampling is performed using a Python multiprocessing Pool instance.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            iterations_per_full_update=iterations_per_full_update,
            iterations_per_quick_update=iterations_per_quick_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs,
        )

        self.logger.debug("Creating Nautilus Search")

    def _fit(self, model: AbstractPriorModel, analysis):
        """
        Fit a model using the search and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that includes the maximum log likelihood instance and full
        set of accepted ssamples of the fit.
        """

        if not isinstance(self.paths, NullPaths):
            checkpoint_exists = os.path.exists(self.checkpoint_file)
        else:
            checkpoint_exists = False

        if checkpoint_exists:
            self.logger.info(
                "Resuming Nautilus non-linear search (previous samples found)."
            )

        else:
            self.logger.info(
                "Starting new Nautilus non-linear search (no previous samples found)."
            )

        if (
            self.config_dict.get("force_x1_cpu")
            or self.kwargs.get("force_x1_cpu")
            or analysis._use_jax
        ):

            fitness = Fitness(
                model=model,
                analysis=analysis,
                paths=self.paths,
                fom_is_log_likelihood=True,
                resample_figure_of_merit=-1.0e99,
                iterations_per_quick_update=self.iterations_per_quick_update,
                use_jax_vmap=True,
                batch_size=self.config_dict_search["n_batch"],
            )

            search_internal = self.fit_x1_cpu(
                fitness=fitness,
                model=model,
                analysis=analysis,
            )

        else:

            fitness = Fitness(
                model=model,
                analysis=analysis,
                paths=self.paths,
                fom_is_log_likelihood=True,
                resample_figure_of_merit=-1.0e99,
                iterations_per_quick_update=self.iterations_per_quick_update
            )

            search_internal = self.fit_multiprocessing(
                fitness=fitness,
                model=model,
                analysis=analysis,
            )

        return search_internal, fitness

    @property
    def sampler_cls(self):
        try:
            from nautilus import Sampler

            return Sampler
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "\n--------------------\n"
                "You are attempting to perform a model-fit using Nautilus. \n\n"
                "However, the optional library Nautilus (https://nautilus-sampler.readthedocs.io/en/stable/index.html) is "
                "not installed.\n\n"
                "Install it via the command `pip install nautilus-sampler==1.0.5`.\n\n"
                "----------------------"
            )

    @property
    def checkpoint_file(self):
        """
        The path to the file used for checkpointing.

        If autofit is not outputting results to hard-disk (e.g. paths is `NullPaths`), this function is bypassed.
        """
        try:
            return self.paths.search_internal_path / "checkpoint.hdf5"
        except TypeError:
            pass

    def fit_x1_cpu(self, fitness, model, analysis):
        """
        Perform the non-linear search, using one CPU core.

        This is used if the likelihood function calls external libraries that cannot be parallelized or use
        threading in a way that conflicts with the parallelization of the non-linear search.

        Parameters
        ----------
        fitness
            The function which takes a model instance and returns its log likelihood via the Analysis class
        model
            The model which maps parameters chosen via the non-linear search (e.g. via the priors or sampling) to
            instances of the model, which are passed to the fitness function.
        analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the search maximizes.
        """

        self.logger.info(
            """
            Running search where parallelization is disabled.
            """
        )

        config_dict = self.config_dict_search
        try:
            config_dict.pop("vectorized")
        except KeyError:
            pass

        search_internal = self.sampler_cls(
            prior=PriorVectorized(model=model),
            likelihood=fitness.call_wrap,
            n_dim=model.prior_count,
            filepath=self.checkpoint_file,
            pool=None,
            vectorized=True,
            **config_dict,
        )

        return self.call_search(search_internal=search_internal, model=model, analysis=analysis, fitness=fitness)

    def fit_multiprocessing(self, fitness, model, analysis):
        """
        Perform the non-linear search, using multiple CPU cores parallelized via Python's multiprocessing module.

        This uses PyAutoFit's sneaky pool class, which allows us to use the multiprocessing module in a way that plays
        nicely with the non-linear search (e.g. exception handling, keyboard interupts, etc.).

        Multiprocessing parallelization can only parallelize across multiple cores on a single device, it cannot be
        distributed across multiple devices or computing nodes. For that, use the `fit_mpi` method.

        Parameters
        ----------
        fitness
            The function which takes a model instance and returns its log likelihood via the Analysis class
        model
            The model which maps parameters chosen via the non-linear search (e.g. via the priors or sampling) to
            instances of the model, which are passed to the fitness function.
        analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the search maximizes.
        """
        search_internal = self.sampler_cls(
            prior=PriorVectorized(model=model),
            likelihood=fitness.call_wrap,
            n_dim=model.prior_count,
            filepath=self.checkpoint_file,
            pool=self.number_of_cores,
            **self.config_dict_search,
        )

        search_internal = self.call_search(
            search_internal=search_internal,
            model=model,
            analysis=analysis,
            fitness=fitness
        )

        return search_internal

    def call_search(self, search_internal, model, analysis, fitness):
        """
        The x1 CPU and multiprocessing searches both call this function to perform the non-linear search.

        This function calls the search a reduced number of times, corresponding to the `iterations_per_full_update` of the
        search. This allows the search to output results on-the-fly, for example writing to the hard-disk the latest
        model and samples.

        It tracks how often to do this update alongside the maximum number of iterations the search will perform.
        This ensures that on-the-fly output is performed at regular intervals and that the search does not perform more
        iterations than the `n_like_max` input variable.

        Parameters
        ----------
        search_internal
            The single CPU or multiprocessing search which is run and performs nested sampling.
        model
            The model which maps parameters chosen via the non-linear search (e.g. via the priors or sampling) to
            instances of the model, which are passed to the fitness function.
        analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the search maximizes.
        """

        finished = False

        minimum_iterations_per_full_updates = 3 * self.config_dict_search["n_live"]

        if self.iterations_per_full_update < minimum_iterations_per_full_updates:

            self.iterations_per_full_update = minimum_iterations_per_full_updates

            logger.info(
                f"""
                The number of iterations_per_full_update is less than 3 times the number of live points, which can cause
                issues where Nautilus loses sampling information due to stopping to output results. The number of
                iterations per update has been increased to 3 times the number of live points, therefore a value
                of {minimum_iterations_per_full_updates}.

                To remove this warning, increase the number of iterations_per_full_update to three or more times the
                number of live points.
                """
            )

        while not finished:

            iterations, total_iterations = self.iterations_from(
                search_internal=search_internal
            )

            config_dict_run = {
                key: value
                for key, value in self.config_dict_run.items()
                if key != "n_like_max"
            }

            search_internal.run(
                **config_dict_run,
                n_like_max=iterations,
            )

            iterations_after_run = self.iterations_from(search_internal=search_internal)[1]

            if (
                    total_iterations == iterations_after_run
                    or iterations_after_run == self.config_dict_run["n_like_max"]
            ):
                finished = True

            if not finished:

                self.perform_update(
                    model=model,
                    analysis=analysis,
                    during_analysis=True,
                    fitness=fitness,
                    search_internal=search_internal
                )

        return search_internal

    def iterations_from(
        self, search_internal
    ) -> Tuple[int, int]:
        """
        Returns the next number of iterations that a dynesty call will use and the total number of iterations
        that have been performed so far.

        This is used so that the `iterations_per_full_update` input leads to on-the-fly output of dynesty results.

        It also ensures dynesty does not perform more samples than the `n_like_max` input variable.

        Parameters
        ----------
        search_internal
            The Dynesty sampler (static or dynamic) which is run and performs nested sampling.

        Returns
        -------
        The next number of iterations that a dynesty run sampling will perform and the total number of iterations
        it has performed so far.
        """

        if isinstance(self.paths, NullPaths):
            n_like_max = self.config_dict_run.get("n_like_max")

            if n_like_max is not None:
                return n_like_max, n_like_max
            return int(1e99), int(1e99)

        try:
            total_iterations = len(search_internal.posterior()[1])
        except ValueError:
            total_iterations = 0

        iterations = total_iterations + self.iterations_per_full_update

        if self.config_dict_run["n_like_max"] is not None:
            if iterations > self.config_dict_run["n_like_max"]:
                iterations = self.config_dict_run["n_like_max"]

        return iterations, total_iterations

    def output_search_internal(self, search_internal):
        """
        Output the sampler results to hard-disk in their internal format.

        The multiprocessing `Pool` object cannot be pickled and thus the sampler cannot be saved to hard-disk. This
        function therefore extracts the necessary information from the sampler and saves it to hard-disk.

        Parameters
        ----------
        sampler
            The nautilus sampler object containing the results of the model-fit.
        """

        pool_l = search_internal.pool_l
        pool_s = search_internal.pool_s

        search_internal.pool_l = None
        search_internal.pool_s = None

        self.paths.save_search_internal(
            obj=search_internal,
        )

        search_internal.pool_l = pool_l
        search_internal.pool_s = pool_s

        try:
            os.remove(self.checkpoint_file)
        except (TypeError, FileNotFoundError):
            pass

    def samples_info_from(self, search_internal=None):
        return {
            "log_evidence": search_internal.evidence(),
            "total_samples": int(search_internal.n_like),
            "total_accepted_samples": int(search_internal.n_like),
            "time": self.timer.time if self.timer else None,
            "number_live_points": int(search_internal.n_live),
        }

    def samples_via_internal_from(
        self, model: AbstractPriorModel, search_internal=None
    ):
        """
        Returns a `Samples` object from the ultranest internal results.

        The samples contain all information on the parameter space sampling (e.g. the parameters,
        log likelihoods, etc.).

        The internal search results are converted from the native format used by the search to lists of values
        (e.g. `parameter_lists`, `log_likelihood_list`).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """

        if search_internal is None:
            search_internal = self.paths.load_search_internal()

        parameters, log_weights, log_likelihoods = search_internal.posterior()

        parameter_lists = parameters.tolist()
        log_likelihood_list = log_likelihoods.tolist()
        weight_list = np.exp(log_weights).tolist()

        log_prior_list = [
            sum(model.log_prior_list_from_vector(vector=vector))
            for vector in parameter_lists
        ]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list,
        )

        return SamplesNest(
            model=model,
            sample_list=sample_list,
            samples_info=self.samples_info_from(search_internal=search_internal),
        )

    @property
    def batch_size(self):
        return self.config_dict_search.get("n_batch")

    @property
    def config_dict(self):
        return conf.instance["non_linear"]["nest"][self.__class__.__name__]

    def config_dict_test_mode_from(self, config_dict : Dict) -> Dict:
        """
        Returns a configuration dictionary for test mode meaning that the sampler terminates as quickly as possible.

        Entries which set the total number of samples of the sampler (e.g. maximum calls, maximum likelihood
        evaluations) are reduced to low values meaning it terminates nearly immediately.

        Parameters
        ----------
        config_dict
            The original configuration dictionary for this sampler which includes entries controlling how fast the
            sampler terminates.

        Returns
        -------
        A configuration dictionary where settings which control the sampler's number of samples are reduced so it
        terminates as quickly as possible.
        """
        return {
            **config_dict,
            "n_like_max": 1,
        }


from __future__ import annotations
from abc import ABC, abstractmethod
from functools import partial
from typing import List, Optional, Generator

import autofit as af


class AggBase(ABC):
    def __init__(self, aggregator: af.Aggregator):
        """
        Base aggregator wrapper, which makes it straight forward to compute generators of instances of objects from
        specific samples of a non-linear search.

        The stadard aggregator makes it straight forward to create instances from the model. However, if there
        are other classes which are generated from the model. but not part of the model itself, creating
        instances of them with samples from the non-linear, via a generator, requires manual code to be written.

        The base aggregator can be used to streamline this process and create a concise API for generating these
        instances.

        This is achieved by overwriting the `object_via_gen_from` method, which is used to create the object from
        the non-linear search samples. This method is then used to create generators of the object from the
        non-linear search samples.

        Parameters
        ----------
        aggregator
            An PyAutoFit aggregator containing the results of non-linear searches performed by PyAutoFit.
        """
        self.aggregator = aggregator

    @abstractmethod
    def object_via_gen_from(
        self, fit: af.Fit, instance: Optional[af.ModelInstance] = None
    ) -> object:
        """
        For example, in the `GalaxiesAgg` object, this function is overwritten such that it creates a `Plane` from a
        `ModelInstance` that contains the galaxies of a sample from a non-linear search.

        Parameters
        ----------
        fit
            A PyAutoFit database Fit object containing the generators of the results of PyAutoGalaxy model-fits.
        instance
            A manual instance that overwrites the max log likelihood instance in fit (e.g. for drawing the instance
            randomly from the PDF).

        Returns
        -------
        Generator
            A generator that creates an object used in the model-fitting process of a non-linear search.
        """

    def max_log_likelihood_gen_from(self) -> Generator:
        """
        Returns a generator using the maximum likelihood instance of a non-linear search.

        This generator creates a list containing the maximum log instance of every result loaded in the aggregator.

        For example, in **PyAutoLens**, by overwriting the `make_gen_from` method this returns a generator
        of `Plane` objects from a PyAutoFit aggregator. This generator then generates a list of the maximum log
        likelihood `Plane` objects for all aggregator results.
        """

        def func_gen(fit: af.Fit) -> Generator:
            return self.object_via_gen_from(fit=fit)

        return self.aggregator.map(func=func_gen)

    def weights_above_gen_from(self, minimum_weight: float) -> List:
        """
        Returns a list of all weights above a minimum weight for every result.

        Parameters
        ----------
        minimum_weight
            The minimum weight of a non-linear sample, such that samples with a weight below this value are discarded
            and not included in the generator.
        """

        def func_gen(fit: af.Fit, minimum_weight: float) -> List[object]:
            samples = fit.samples

            weight_list = []

            for sample in samples.sample_list:
                if sample.weight > minimum_weight:
                    weight_list.append(sample.weight)

            return weight_list

        func = partial(func_gen, minimum_weight=minimum_weight)

        return self.aggregator.map(func=func)

    def all_above_weight_gen_from(self, minimum_weight: float) -> Generator:
        """
        Returns a generator which for every result generates a list of objects whose parameter values are all those
        in the non-linear search with a weight about an input `minimum_weight` value. This enables straight forward
        error estimation.

        This generator creates lists containing instances whose non-linear sample weight are above the value of
        `minimum_weight`. For example, if the aggregator contains 10 results and each result has 100 samples above the
        `minimum_weight`, a list of 10 entries will be returned, where each entry in this list contains 100 object's
        paired with each non-linear sample.

        For example, in **PyAutoLens**, by overwriting the `make_gen_from` method this returns a generator
        of `Plane` objects from a PyAutoFit aggregator. This generator then generates lists of `Plane` objects
        corresponding to all non-linear search samples above the `minimum_weight`.

        Parameters
        ----------
        minimum_weight
            The minimum weight of a non-linear sample, such that samples with a weight below this value are discarded
            and not included in the generator.
        """

        def func_gen(fit: af.Fit, minimum_weight: float) -> List[object]:
            samples = fit.samples

            all_above_weight_list = []

            for sample in samples.sample_list:
                if sample.weight > minimum_weight:
                    instance = sample.instance_for_model(model=samples.model)

                    all_above_weight_list.append(
                        self.object_via_gen_from(fit=fit, instance=instance)
                    )

            return all_above_weight_list

        func = partial(func_gen, minimum_weight=minimum_weight)

        return self.aggregator.map(func=func)

    def randomly_drawn_via_pdf_gen_from(self, total_samples: int):
        """
        Returns a generator which for every result generates a list of objects whose parameter values are drawn
        randomly from the PDF. This enables straight forward error estimation.

        This generator creates lists containing instances that are drawn randomly from the PDF for every result loaded
        in the aggregator. For example, the aggregator contains 10 results and if `total_samples=100`, a list of 10
        entries will be returned, where each entry in this list contains 100 object's paired with non-linear samples
        randomly drawn from the PDF.

        For example, in **PyAutoLens**, by overwriting the `make_gen_from` method this returns a generator
        of `Plane` objects from a PyAutoFit aggregator. This generator then generates lists of `Plane` objects
        corresponding to non-linear search samples randomly drawn from the PDF.

        Parameters
        ----------
        total_samples
            The total number of non-linear search samples that should be randomly drawn from the PDF.
        """

        def func_gen(fit: af.Fit, total_samples: int) -> List[object]:
            samples = fit.samples

            return [
                self.object_via_gen_from(
                    fit=fit,
                    instance=samples.draw_randomly_via_pdf(),
                )
                for i in range(total_samples)
            ]

        func = partial(func_gen, total_samples=total_samples)

        return self.aggregator.map(func=func)



import dill

from .model import Object
from ..sqlalchemy_ import sa


class Function(Object):
    """
    A function or callable
    """

    __tablename__ = "function"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )
    serialised_function = sa.Column(sa.String)

    @property
    def function(self) -> callable:
        """
        The function is stored as a serialised dill string
        """
        return dill.loads(self.serialised_function)

    @function.setter
    def function(self, function):
        self.serialised_function = dill.dumps(function)

    __mapper_args__ = {"polymorphic_identity": "function"}

    @classmethod
    def _from_object(cls, source):
        return cls(function=source)

    def __call__(self):
        return self.function



import numpy as np

from autoconf.class_path import get_class_path, get_class
from .model import Object
from ..sqlalchemy_ import sa


class Array(Object):
    """
    A serialised numpy array
    """

    __tablename__ = "array"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "array"}

    bytes = sa.Column(sa.LargeBinary)
    _dtype = sa.Column(sa.String)
    _shape = sa.Column(sa.String)

    fit_id = sa.Column(sa.String, sa.ForeignKey("fit.id"))
    fit = sa.orm.relationship(
        "Fit",
        uselist=False,
        foreign_keys=[fit_id],
        back_populates="arrays",
    )

    @property
    def shape(self):
        return tuple(map(int, self._shape.split(",")))

    @shape.setter
    def shape(self, shape):
        self._shape = ",".join(map(str, shape))

    @property
    def array(self):
        return np.frombuffer(
            self.bytes,
            dtype=self.dtype,
        ).reshape(self.shape)

    @array.setter
    def array(self, array):
        self._dtype = get_class_path(getattr(np, array.dtype.name))
        self.shape = array.shape
        self.bytes = array.tobytes()

    @property
    def dtype(self):
        return get_class(self._dtype)

    @property
    def value(self):
        return self.array

    def __call__(self, *args, **kwargs):
        return self.value


class Fits(Object):
    """
    A serialised astropy.io.fits.HDUList
    """

    __tablename__ = "fits"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "fits",
    }

    hdus = sa.orm.relationship(
        "HDU",
        back_populates="fits",
        foreign_keys="HDU.fits_id",
    )

    fit_id = sa.Column(sa.String, sa.ForeignKey("fit.id"))
    fit = sa.orm.relationship(
        "Fit",
        uselist=False,
        foreign_keys=[fit_id],
        back_populates="fits",
    )

    @property
    def hdu_list(self):
        from astropy.io import fits

        return fits.HDUList([hdu.hdu for hdu in self.hdus])

    @hdu_list.setter
    def hdu_list(self, hdu_list):
        self.hdus = [HDU(hdu=hdu) for hdu in hdu_list]

    @property
    def value(self):
        return self.hdu_list


class HDU(Array):
    """
    A serialised astropy.io.fits.PrimaryHDU
    """

    __tablename__ = "hdu"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("array.id"),
        primary_key=True,
        index=True,
    )

    _header = sa.Column(sa.String)

    __mapper_args__ = {
        "polymorphic_identity": "hdu",
    }

    is_primary = sa.Column(sa.Boolean)

    fit = sa.orm.relationship(
        "Fit",
        uselist=False,
        foreign_keys=[Array.fit_id],
        back_populates="hdus",
    )

    fits_id = sa.Column(
        sa.Integer,
        sa.ForeignKey("fits.id"),
    )
    fits = sa.orm.relationship(
        "Fits",
        uselist=False,
        foreign_keys=[fits_id],
        back_populates="hdus",
    )

    @property
    def header(self):
        """
        The header of the HDU
        """
        from astropy.io.fits import Header

        return Header.fromstring(self._header)

    @header.setter
    def header(self, header):
        self._header = header.tostring()

    @property
    def hdu(self):
        from astropy.io import fits

        type_ = fits.PrimaryHDU if self.is_primary else fits.ImageHDU

        return type_(
            self.array,
            self.header,
        )

    @hdu.setter
    def hdu(self, hdu):
        from astropy.io import fits

        self.is_primary = isinstance(hdu, fits.PrimaryHDU)
        self.array = hdu.data
        self.header = hdu.header

    @property
    def value(self):
        return self.hdu

    @property
    def dtype(self):
        return np.dtype(">f8")



from autofit.mapper.prior_model.collection import Collection
from autofit.non_linear.samples.sample import Sample
from autofit.non_linear.samples.summary import SamplesSummary


class MockSamplesSummary(SamplesSummary):
    def __init__(
        self,
        model=None,
        max_log_likelihood_sample=None,
        median_pdf_sample=None,
        log_evidence=None,
        max_log_likelihood_instance=None,
        prior_means=None,
        **kwargs,
    ):
        super().__init__(
            model=model,
            max_log_likelihood_sample=max_log_likelihood_sample,
            median_pdf_sample=median_pdf_sample,
            log_evidence=log_evidence,
        )

        self._max_log_likelihood_instance = max_log_likelihood_instance
        self._prior_means = prior_means
        self._kwargs = {path: 1.0 for path in self.model.paths} if self.model else {}

    @property
    def max_log_likelihood_sample(self):
        if self._max_log_likelihood_sample is not None:
            return self._max_log_likelihood_sample

        return Sample(
            log_likelihood=1.0,
            log_prior=0.0,
            weight=0.0,
            kwargs=self._kwargs,
        )

    @property
    def median_pdf_sample(self):
        if self._median_pdf_sample is not None:
            return self._median_pdf_sample

        return Sample(
            log_likelihood=1.0,
            log_prior=0.0,
            weight=0.0,
            kwargs=self._kwargs,
        )

    def max_log_likelihood(self, as_instance: bool = True):
        if self._max_log_likelihood_instance is None:
            try:
                return super().max_log_likelihood(as_instance=as_instance)
            except (KeyError, AttributeError):
                pass

        return self._max_log_likelihood_instance

    @property
    def prior_means(self):
        if self._prior_means is None:
            return super().prior_means

        return self._prior_means

    @classmethod
    def default(cls):
        return MockSamplesSummary(
            model=Collection(),
        )



import csv
import inspect
from copy import copy
from pathlib import Path
from typing import List, Tuple, Union

from autoconf.class_path import get_class_path
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.tools.util import split_paths


class Sample:
    def __init__(
        self,
        log_likelihood: float,
        log_prior: float,
        weight: float,
        kwargs=None,
    ):
        """
        One sample taken during a search

        Parameters
        ----------
        log_likelihood
            The likelihood associated with this instance
        log_prior
            A logarithmic prior of the instance
        weight
            The weight this sample contributes to the PDF.
        kwargs
            Dictionary mapping model paths to values for the sample
        """
        self.log_likelihood = float(log_likelihood)
        self.log_prior = log_prior
        self.weight = weight
        self.kwargs = {
            tuple(key.split(".")) if isinstance(key, str) and "." in key else key: value
            for key, value in (kwargs or dict()).items()
        }

    def dict(self):
        return {
            "type": "instance",
            "class_path": get_class_path(type(self)),
            "arguments": {
                "log_likelihood": self.log_likelihood,
                "log_prior": self.log_prior,
                "weight": self.weight,
                "kwargs": {
                    "type": "dict",
                    "arguments": {
                        ".".join(key) if isinstance(key, tuple) else key: value
                        for key, value in self.kwargs.items()
                    },
                },
            },
        }

    def model_dict(self) -> dict:
        """
        A dictionary mapping model paths to values for the sample
        """
        model_dict = dict()
        for key, value in self.kwargs.items():
            current = model_dict
            if not isinstance(key, tuple):
                key = (key,)
            for part in key[:-1]:
                if part not in current:
                    current[part] = dict()
                current = current[part]
            current[key[-1]] = value
        return model_dict

    @property
    def log_posterior(self) -> float:
        """
        Compute the posterior
        """
        return self.log_likelihood + self.log_prior

    def parameter_lists_for_model(self, model: AbstractPriorModel) -> List[float]:
        """
        Values for instantiating a model, in the same order as priors
        from the model.

        Parameters
        ----------
        model
            The model from which this was a sample

        Returns
        -------
        A list of physical values
        """
        if self.is_path_kwargs:
            paths = model.all_paths
        else:
            paths = model.all_names

        return self.parameter_lists_for_paths(paths)

    def parameter_lists_for_paths(self, paths):
        result = list()
        for keys in paths:
            is_found = False
            for key in keys:
                if key in self.kwargs:
                    result.append(self.kwargs[key])
                    is_found = True
                    break
            if not is_found:
                raise KeyError(
                    f"Could not find any of the following keys in kwargs {keys}"
                )
        return result

    @property
    def is_path_kwargs(self) -> bool:
        """
        Are the keys in the kwargs dictionary tuples? If they
        are this indicates that they are explicit paths through
        the model.
        """
        for key in self.kwargs:
            return isinstance(key, tuple)
        return False

    def subsample(self, path_map):
        arg_dict = {}
        for paths, new_path in path_map.items():
            for path in paths:
                if path in self.kwargs:
                    arg_dict[new_path] = self.kwargs[path]
                    break
            else:
                raise KeyError(f"No path from {paths} in sample")

        return type(self)(
            log_likelihood=self.log_likelihood,
            log_prior=self.log_prior,
            weight=self.weight,
            kwargs=arg_dict,
        )

    @classmethod
    def from_lists(
        cls,
        model: AbstractPriorModel,
        parameter_lists: List[List[float]],
        log_likelihood_list: List[float],
        log_prior_list: List[float],
        weight_list: List[float],
    ) -> List["Sample"]:
        """
        Convenience method to create a list of samples from lists of contained values
        """
        samples = list()

        paths = model.unique_prior_paths

        for params, log_likelihood, log_prior, weight in zip(
            parameter_lists, log_likelihood_list, log_prior_list, weight_list
        ):
            arg_dict = {t: param for t, param in zip(paths, params)}

            samples.append(
                cls(
                    log_likelihood=log_likelihood,
                    log_prior=log_prior,
                    weight=weight,
                    kwargs=arg_dict,
                )
            )
        return samples

    def instance_for_model(
        self,
        model: AbstractPriorModel,
        ignore_assertions: bool = False,
    ):
        """
        Create an instance from this sample for a model

        Parameters
        ----------
        model
            The model the this sample was taken from
        ignore_assertions
            If True, do not check that the instance is valid

        Returns
        -------
        The instance corresponding to this sample
        """
        try:
            if self.is_path_kwargs:
                return model.instance_from_path_arguments(
                    self.kwargs,
                    ignore_assertions=ignore_assertions,
                )
            else:
                return model.instance_from_prior_name_arguments(
                    self.kwargs,
                    ignore_assertions=ignore_assertions,
                )

        except KeyError:
            return model.instance_from_vector(
                self.parameter_lists_for_model(model),
                ignore_assertions=ignore_assertions,
            )

    @split_paths
    def with_paths(self, paths: List[Tuple[str, ...]]) -> "Sample":
        """
        Create a copy of this object retaining only the kwargs for which
        there is a matching path in paths.

        Parameters
        ----------
        paths
            A list of paths for which attributes should be retained.

        Returns
        -------
        A reduced sample
        """
        with_paths = copy(self)
        with_paths.kwargs = {
            key: value
            for key, value in self.kwargs.items()
            if any(
                all(first == second for first, second in zip(key, path))
                for path in paths
            )
        }
        return with_paths

    @split_paths
    def without_paths(self, paths: List[Tuple[str, ...]]) -> "Sample":
        """
        Create a copy of this object retaining only the kwargs for which
        there is no matching path in paths.

        Parameters
        ----------
        paths
            A list of paths for which attributes should be removed.

        Returns
        -------
        A reduced sample
        """
        without_paths = copy(self)
        without_paths.kwargs = {
            key: value
            for key, value in self.kwargs.items()
            if not any(
                all(first == second for first, second in zip(key, path))
                for path in paths
            )
        }
        return without_paths


sample_args = set(inspect.getfullargspec(Sample.__init__).args)


def samples_from_iterator(iterator):
    samples = list()

    headers = next(iterator)
    headers = [header.strip() for header in headers]
    for row in iterator:
        d = {header: float(value) for header, value in zip(headers, row)}

        samples.append(
            Sample(
                **{key: value for key, value in d.items() if key in sample_args},
                kwargs={
                    key: value
                    for key, value in d.items()
                    if key not in (sample_args | {"log_posterior"})
                },
            )
        )
    return samples


def load_from_table(filename: Union[str, Path]) -> List[Sample]:
    """
    Load samples from a table

    Parameters
    ----------
    filename
        The path to a CSV file

    Returns
    -------
    A list of samples, one for each row in the CSV
    """
    with open(filename, "r+", newline="") as f:
        reader = csv.reader(f)
        return samples_from_iterator(reader)



import copy
import logging
from functools import wraps
from typing import Optional, Union, Tuple, List, Iterable, Type, Dict

from autofit.mapper.model_object import ModelObject
from autofit.mapper.prior_model.recursion import DynamicRecursionCache

logger = logging.getLogger(__name__)


def frozen_cache(func):
    """
    Decorator that caches results from function calls when
    a model is frozen.

    Value is cached by function name, instance and arguments.

    Parameters
    ----------
    func
        Some function attached to a freezable, hashable object
        that takes hashable arguments

    Returns
    -------
    Function with cache
    """

    @wraps(func)
    def cache(self, *args, **kwargs):
        if hasattr(self, "_is_frozen") and self._is_frozen:
            key = (
                func.__name__,
                self,
                *args,
            ) + tuple(kwargs.items())

            if key not in self._frozen_cache:
                self._frozen_cache[key] = func(self, *args, **kwargs)
            return self._frozen_cache[key]
        return func(self, *args, **kwargs)

    return cache


def assert_not_frozen(func):
    """
    Decorator that asserts a function is not called when an object
    is frozen. For example, it should not be possible to set an
    attribute on a frozen model as that might invalidate the results
    in the cache.

    Parameters
    ----------
    func
        Some function

    Raises
    ------
    AssertionError
        If the function is called when the object is frozen
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        string_args = list(filter(lambda arg: isinstance(arg, str), args))
        if (
            "_is_frozen" not in string_args
            and "_frozen_cache" not in string_args
            and hasattr(self, "_is_frozen")
            and self._is_frozen
        ):
            raise AssertionError("Frozen models cannot be modified")
        return func(self, *args, **kwargs)

    return wrapper


class AbstractModel(ModelObject):
    def __init__(self, label=None, id_=None):
        self._is_frozen = False
        self._frozen_cache = dict()
        super().__init__(label=label, id_=id_)

    def __getstate__(self):
        return {
            key: value for key, value in self.__dict__.items() if key != "_frozen_cache"
        }

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._frozen_cache = {}

    def freeze(self):
        """
        Freeze this object.

        A frozen object caches results for some function calls
        and does not allow its state to be modified.
        """
        logger.debug("Freezing model")
        tuples = self.direct_tuples_with_type(AbstractModel)
        for _, model in tuples:
            if model is not self:
                model.freeze()
        self._is_frozen = True

    def unfreeze(self):
        """
        Unfreeze this object. Allows modification and removes
        caches associated with some functions.
        """
        logger.debug("Thawing model")
        self._is_frozen = False
        tuples = self.direct_tuples_with_type(AbstractModel)
        for _, model in tuples:
            if model is not self:
                model.unfreeze()
        self._frozen_cache = dict()

    def __add__(self, other):
        instance = self.__class__()

        def add_items(item_dict):
            for key, value in item_dict.items():
                if isinstance(value, list) and hasattr(instance, key):
                    setattr(instance, key, getattr(instance, key) + value)
                else:
                    setattr(instance, key, value)

        add_items(self.__dict__)
        add_items(other.__dict__)
        return instance

    def copy(self):
        """
        Create a copy of the model. All priors remain equivalent - i.e. two
        copies of a model in a collection has the same prior count as a single
        model.
        """
        return copy.deepcopy(self)

    def object_for_path(
        self, path: Iterable[Union[str, int, type]]
    ) -> Union[object, List]:
        """
        Get the object at a given path.

        The path describes the location of some object in the model.

        String entries get an attribute.
        Int entries index an attribute.
        Type entries product a new ModelInstance which collates all of the instances
        of a given type in the path.

        Parameters
        ----------
        path
            A tuple describing the path to an object in the model tree

        Returns
        -------
        An object or Instance collating a collection of objects with a given type.
        """
        instance = self
        for name in path:
            if isinstance(name, int):
                instance = instance[name]
            elif isinstance(name, type):
                from autofit.mapper.prior_model.prior_model import Model

                instances = [
                    instance
                    for _, instance in self.path_instance_tuples_for_class(name)
                ]
                instances += [
                    instance
                    for _, instance in self.path_instance_tuples_for_class(Model)
                    if issubclass(instance.cls, name)
                ]
                instance = ModelInstance(instances)
            else:
                instance = getattr(instance, name)
        return instance

    @frozen_cache
    def path_instance_tuples_for_class(
        self,
        cls: Union[Tuple, Type],
        ignore_class: Optional[type] = None,
        ignore_children: bool = True,
    ):
        """
        Tuples containing the path tuple and instance for every instance of the class
        in the model tree.

        Parameters
        ----------
        ignore_class
            Children of instances of this class are ignored
        ignore_children
            If true do not continue to recurse the children of an object once found
        cls
            The type to find instances of

        Returns
        -------
        path_instance_tuples: [((str,), object)]
            Tuples containing the path to and instance of objects of the given type.
        """
        return path_instances_of_class(
            self, cls, ignore_class=ignore_class, ignore_children=ignore_children
        )

    @frozen_cache
    def direct_tuples_with_type(self, class_type):
        return list(
            filter(
                lambda t: t[0] != "id"
                and not t[0].startswith("_")
                and isinstance(t[1], class_type),
                self.__dict__.items(),
            )
        )

    @frozen_cache
    def models_with_type(
        self,
        cls: Union[Type, Tuple[Type, ...]],
        include_zero_dimension=False,
    ) -> List["AbstractModel"]:
        """
        Return all models of a given type in the model tree.

        Parameters
        ----------
        cls
            The type to find instances of
        include_zero_dimension
            If true, include models with zero dimensions

        Returns
        -------
        A list of models of the given type
        """
        # noinspection PyTypeChecker
        return [
            t[1]
            for t in self.model_tuples_with_type(
                cls, include_zero_dimension=include_zero_dimension
            )
        ]

    @frozen_cache
    def model_tuples_with_type(
        self, cls: Union[Type, Tuple[Type, ...]], include_zero_dimension=False
    ):
        """
        All models of the class in this model which have at least
        one free parameter, recursively.

        Parameters
        ----------
        cls
            The type of the model
        include_zero_dimension
            If true, include models with 0 free parameters

        Returns
        -------
        Models with free parameters
        """
        from .prior_model.prior_model import Model

        return [
            (path, model)
            for path, model in self.attribute_tuples_with_type(
                Model, ignore_children=False
            )
            if issubclass(model.cls, cls)
            and (include_zero_dimension or model.prior_count > 0)
        ]

    @frozen_cache
    def attribute_tuples_with_type(
        self,
        class_type,
        ignore_class=None,
        ignore_children=True,
    ) -> List[tuple]:
        """
        Tuples describing the name and instance for attributes in the model
        with a given type, recursively.

        Parameters
        ----------
        ignore_children
            If True then recursion stops at instances with the type
        class_type
            The type of the objects to find
        ignore_class
            Any classes which should not be recursively searched

        Returns
        -------
        Tuples containing the name and instance of each attribute with the type
        """
        return [
            (path[-1] if len(path) > 0 else "", value)
            for path, value in self.path_instance_tuples_for_class(
                class_type,
                ignore_class=ignore_class,
                ignore_children=ignore_children,
            )
        ]


@DynamicRecursionCache()
def path_instances_of_class(
    obj,
    cls: type,
    ignore_class: Optional[Union[type, Tuple[type]]] = None,
    ignore_children: bool = False,
):
    """
    Recursively search the object for instances of a given class

    Parameters
    ----------
    obj
        The object to recursively search
    cls
        The type to search for
    ignore_class
        A type or tuple of classes to skip
    ignore_children
        If true stop recursion at found objects

    Returns
    -------
    instance of type
    """
    if ignore_class is not None and isinstance(obj, ignore_class):
        return []

    results = []
    if isinstance(obj, cls):
        results.append((tuple(), obj))
        if ignore_children:
            return results

    if isinstance(obj, list):
        for i, item in enumerate(obj):
            for path, instance in path_instances_of_class(
                item, cls, ignore_class=ignore_class, ignore_children=ignore_children
            ):
                results.append(((i,) + path, instance))
        return results

    try:
        from autofit.mapper.prior_model.annotation import AnnotationPriorModel

        if isinstance(obj, dict):
            d = obj
        else:
            d = obj.__dict__

        for key, value in d.items():
            if key.startswith("_"):
                continue
            for item in path_instances_of_class(
                value, cls, ignore_class=ignore_class, ignore_children=ignore_children
            ):
                if isinstance(value, AnnotationPriorModel):
                    path = (key,)
                else:
                    path = (key, *item[0])
                results.append((path, item[1]))
        return results
    except (AttributeError, TypeError):
        return results


class ModelInstance(AbstractModel):
    """
    An instance of a Collection or Model. This is created by optimisers and correspond
    to a point in the parameter space.

    @DynamicAttrs
    """

    __dictable_type__ = "instance"

    def __init__(self, child_items: Optional[Union[List, Dict]] = None, id_=None):
        """
        An instance of a Collection or Model. This is created by optimisers and correspond
        to a point in the parameter space.

        Parameters
        ----------
        child_items
            The child items of the instance. This can be a list or dict.

            If a list, the items are assigned to the instance in order.
            If a dict, the items are assigned to the instance by key and accessed by attribute.
        """
        super().__init__()
        self.child_items = child_items
        self.id = id_

    def __eq__(self, other):
        try:
            return self.__dict__ == other.__dict__
        except AttributeError:
            return False

    def __getitem__(self, item):
        if isinstance(item, int):
            return list(self.values())[item]
        if isinstance(item, slice):
            return ModelInstance(list(self.values())[item])
        return self.__dict__[item]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    @property
    def child_items(self):
        return self.dict

    @child_items.setter
    def child_items(self, child_items):
        if isinstance(child_items, list):
            for i, item in enumerate(child_items):
                self[i] = item
        if isinstance(child_items, dict):
            for key, value in child_items.items():
                self[key] = value

    def items(self):
        return self.dict.items()

    def __hash__(self):
        return self.id

    @property
    def dict(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in ("id", "component_number", "item_number")
            and not (isinstance(key, str) and key.startswith("_"))
        }

    def tree_flatten(self) -> Tuple[List, Tuple]:
        """
        Flatten the instance into a PyTree
        """
        keys, values = zip(*self.dict.items())
        return values, (
            *keys,
            self.id,
        )

    @classmethod
    def tree_unflatten(
        cls,
        aux_data: Tuple,
        children: List,
    ):
        """
        Create an instance from a flattened PyTree

        Parameters
        ----------
        aux_data
            Auxiliary information that remains unchanged including
            the keys of the dict
        children
            Child objects subject to change

        Returns
        -------
        An instance of this class
        """
        *keys, id_ = aux_data

        instance = cls(id_=id_)

        for key, value in zip(keys, children):
            instance[key] = value
        return instance

    def values(self):
        return self.dict.values()

    def __len__(self):
        return len(self.values())

    def as_model(
        self,
        model_classes: Union[type, Iterable[type]] = tuple(),
        excluded_classes: Union[type, Iterable[type]] = tuple(),
    ):
        """
        Convert this instance to a model

        Parameters
        ----------
        model_classes
            The classes to convert to models
        excluded_classes
            The classes to exclude from conversion

        Returns
        -------
        A model
        """

        from autofit.mapper.prior_model.abstract import AbstractPriorModel

        return AbstractPriorModel.from_instance(
            self,
            model_classes,
            excluded_classes,
        )



from .abstract import AbstractInterpolator
from autofit.interpolator.linear_relationship import LinearRelationship


class LinearInterpolator(AbstractInterpolator):
    """
    Assume all attributes have a linear relationship with time
    """

    @staticmethod
    def _relationship(x, y):

        from scipy.stats import linregress

        slope, intercept, r, p, std_err = linregress(x, y)

        return LinearRelationship(slope, intercept)



from typing import Optional

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from .analysis import Analysis
from ... import SamplesSummary, AbstractPaths, SamplesPDF


class ModelAnalysis(Analysis):
    def __init__(self, analysis: Analysis, model: AbstractPriorModel, use_jax : bool = False):
        """
        Comprises a model and an analysis that can be applied to instances of that model.

        Parameters
        ----------
        analysis
        model
        """
        super().__init__(use_jax=use_jax)

        self.analysis = analysis
        self.model = model

    def __getattr__(self, item):
        if item in ("__getstate__", "__setstate__"):
            raise AttributeError(item)
        return getattr(self.analysis, item)

    def log_likelihood_function(self, instance):
        return self.analysis.log_likelihood_function(instance)

    def make_result(
        self,
        samples_summary: SamplesSummary,
        paths: AbstractPaths,
        samples: Optional[SamplesPDF] = None,
        search_internal: Optional[object] = None,
        analysis: Optional[object] = None,
    ):
        """
        Return the correct type of result by calling the underlying analysis.
        """
        try:
            return self.analysis.make_result(
                samples_summary=samples_summary,
                paths=paths,
                samples=samples,
                search_internal=search_internal,
            )
        except TypeError:
            raise



from typing import Set

from autofit.database.query.condition import AbstractCondition, Table, info_table
from autofit.database.query.query import AbstractQuery


class InfoQueryCondition(AbstractCondition):
    def __init__(self, key: str, value: str):
        """
        Query some item stored in the fit info dictionary

        Parameters
        ----------
        key
        value
        """
        self.key = key
        self.value = value

    @property
    def tables(self) -> Set[Table]:
        return {info_table}

    def __str__(self):
        return f"key = '{self.key}' AND value = '{self.value}'"


class InfoQuery(AbstractQuery):
    def __init__(self, key, value):
        super().__init__(
            InfoQueryCondition(
                key=key,
                value=value
            )
        )

    @property
    def fit_query(self) -> str:
        return f"SELECT fit_id FROM info WHERE {self.condition}"


class InfoField:
    def __init__(self, key):
        self.key = key

    def __eq__(self, other):
        return InfoQuery(
            self.key,
            other
        )


class AnonymousInfo:
    def __getitem__(self, item):
        return InfoField(item)



from abc import ABC, abstractmethod
from typing import Tuple, Iterator
from typing import Union

import numpy as np

from autoconf import cached_property


class MessageInterface(ABC):
    """
    Common base class for base and transformed messages
    """

    log_base_measure: float
    log_norm: float
    id: int

    @property
    @abstractmethod
    def broadcast(self):
        pass

    @property
    def shape(self) -> Tuple[int, ...]:

        # JAX behaviour
        if isinstance(self.broadcast, list):
            return ()

        return self.broadcast.shape

    @property
    def size(self) -> int:
        return self.broadcast.size

    @property
    def ndim(self) -> int:
        return self.broadcast.ndim

    def __eq__(self, other):
        return self.id == other.id

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return np.exp(self.logpdf(x))

    def logpdf(self, x: Union[np.ndarray, float], xp=np) -> np.ndarray:

        eta = self._broadcast_natural_parameters(x, xp=xp)
        t = self.to_canonical_form(x, xp=xp)
        log_base = self.calc_log_base_measure(x, xp=xp)

        return self.natural_logpdf(eta, t, log_base, self.log_partition(xp=xp), xp=xp)

    def _broadcast_natural_parameters(self, x, xp=np):
        shape = xp.shape(x)
        if shape == self.shape:
            return self.natural_parameters(xp=xp)
        elif shape[1:] == self.shape:
            return self.natural_parameters(xp=xp)[:, None, ...]
        else:
            raise ValueError(
                f"shape of passed value {shape} does not "
                f"match message shape {self.shape}"
            )

    @abstractmethod
    def natural_parameters(self, xp=np) -> np.ndarray:
        pass

    @staticmethod
    @abstractmethod
    def to_canonical_form(x: Union[np.ndarray, float], xp=np) -> np.ndarray:
        pass

    @classmethod
    def calc_log_base_measure(cls, x, xp=np):
        return cls.log_base_measure

    @abstractmethod
    def log_partition(self, xp=np) -> np.ndarray:
        pass

    @classmethod
    def natural_logpdf(cls, eta, t, log_base, log_partition, xp=np):
        eta_t = xp.multiply(eta, t).sum(0)
        return xp.nan_to_num(log_base + eta_t - log_partition, nan=-xp.inf)

    def numerical_logpdf_gradient(
        self, x: np.ndarray, eps: float = 1e-6
    ) -> Tuple[np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            logl0 = self.logpdf(x0)
            if self.multivariate:
                grad_logl = np.empty(logl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(logl0.ndim))
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0)
                        grad_logl[sl + it.multi_index] = (logl - logl0) / eps
                        xv -= eps
            else:
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0).sum()  # type: ignore
                        grad_logl[it.multi_index] = (logl - l0) / eps
                        xv -= eps
        else:
            logl0 = self.logpdf(x)
            grad_logl = (self.logpdf(x + eps) - logl0) / eps

        return logl0, grad_logl

    logpdf_gradient = numerical_logpdf_gradient

    @property
    @abstractmethod
    def multivariate(self):
        pass

    def numerical_logpdf_gradient_hessian(
        self, x: np.ndarray, eps: float = 1e-6
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            if self.multivariate:
                logl0, gradl0 = self.numerical_logpdf_gradient(x0)
                hess_logl = np.empty(gradl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(gradl0.ndim))
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        _, gradl = self.numerical_logpdf_gradient(x0)
                        hess_logl[sl + it.multi_index] = (gradl - gradl0) / eps
                        xv -= eps
            else:
                logl0 = self.logpdf(x0)
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                hess_logl = np.empty_like(x0)
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        l1 = self.logpdf(x0).sum()
                        xv -= 2 * eps
                        l2 = self.logpdf(x0).sum()
                        g1 = (l1 - l0) / eps
                        g2 = (l0 - l2) / eps
                        grad_logl[it.multi_index] = g1
                        hess_logl[it.multi_index] = (g1 - g2) / eps
                        xv += eps

                gradl0 = grad_logl
        else:
            logl0 = self.logpdf(x)
            logl1 = self.logpdf(x + eps)
            logl2 = self.logpdf(x - eps)
            gradl0 = (logl1 - logl0) / eps
            gradl1 = (logl0 - logl2) / eps
            hess_logl = (gradl0 - gradl1) / eps

        return logl0, gradl0, hess_logl

    logpdf_gradient_hessian = numerical_logpdf_gradient_hessian

    @staticmethod
    def _iter_dists(dists) -> Iterator[Union["MessageInterface", float]]:
        for elem in dists:
            if isinstance(elem, MessageInterface):
                yield elem
            elif np.isscalar(elem):
                yield elem
            else:
                for dist in elem:
                    yield dist

    @abstractmethod
    def check_support(self) -> np.ndarray:
        pass

    def check_finite(self) -> np.ndarray:
        return np.isfinite(self.natural_parameters()).all(0)

    def check_valid(self) -> np.ndarray:
        return self.check_finite() & self.check_support()

    @cached_property
    def is_valid(self) -> Union[np.ndarray, np.bool_]:
        return np.all(self.check_finite()) and np.all(self.check_support())

    @abstractmethod
    def update_invalid(self, other: "MessageInterface") -> "MessageInterface":
        pass

    def sum_natural_parameters(self, *dists: "MessageInterface") -> "MessageInterface":
        """return the unnormalised result of multiplying the pdf
        of this distribution with another distribution of the same
        type
        """
        new_params = sum(
            (
                dist.natural_parameters()
                for dist in self._iter_dists(dists)
                if isinstance(dist, MessageInterface)
            ),
            self.natural_parameters(),
        )
        return self.from_natural_parameters(
            new_params,
            id_=self.id,
        )

    def sub_natural_parameters(self, other: "MessageInterface") -> "MessageInterface":
        """return the unnormalised result of dividing the pdf
        of this distribution with another distribution of the same
        type"""
        log_norm = self.log_norm - other.log_norm
        new_params = self.natural_parameters() - other.natural_parameters()
        return self.from_natural_parameters(
            new_params,
            log_norm=log_norm,
            id_=self.id,
        )

    @abstractmethod
    def from_natural_parameters(self, new_params, **kwargs):
        pass

    _multiply = sum_natural_parameters
    _divide = sub_natural_parameters



from pathlib import Path
from typing import List
import logging

from sqlalchemy.orm import Session

from autofit.database.model.fit import Fit
from autofit.text.formatter import write_table

logger = logging.getLogger(__name__)


class Info:
    def __init__(self, session: Session):
        """
        Write info describing the contents of a database file
        in terms of the fits it contains.

        Parameters
        ----------
        session
            A database session
        """
        self.session = session

    @property
    def fits(self) -> List[Fit]:
        """
        A list of objects representing all the fits in the database.
        """
        return self.session.query(Fit).all()

    @property
    def headers(self) -> List[str]:
        """
        A list of strings describing the columns of the table.
        """
        return [
            "unique_id",
            "name",
            "unique_tag",
            "total_free_parameters",
            "is_complete",
        ]

    @property
    def rows(self) -> List[List[str]]:
        """
        A list of lists of strings describing the rows of the table.
        """
        return [
            [
                fit.id,
                fit.name,
                fit.unique_tag,
                fit.total_parameters,
                fit.is_complete,
            ]
            for fit in self.fits
        ]

    @property
    def path(self) -> Path:
        """
        The path to the file where the info will be written.
        """
        return Path(self.session.bind.url.database).with_suffix(".info")

    def write(self):
        """
        Write the info to a file.
        """
        try:
            write_table(
                self.headers,
                self.rows,
                str(self.path),
            )
        except TypeError:
            logger.warning("Failed to write database info.")



import logging
import math
from itertools import count
from multiprocessing import Process
from multiprocessing import Queue

from autofit.non_linear.parallel.sneaky import StopCommand
from autofit.non_linear.paths.abstract import AbstractPaths

logger = logging.getLogger(__name__)


class AnalysisProcess(Process):
    _id = count()

    def __init__(self, analyses):
        """
        A process that performs one or more analyses on each
        instance passed to it

        Parameters
        ----------
        analyses
            A list of analyses this process performs
        """
        super().__init__()
        self.analyses = analyses
        self.instance_queue = Queue()
        self.queue = Queue()

        self.name = f"analysis_process_{next(self._id)}"

        self.logger = logging.getLogger(self.name)
        self.logger.info("Created")

    def _run(self):
        """
        Private run function permitting process to be stopped using a return
        when a StopCommand is passed
        """
        while True:
            instance = self.instance_queue.get()
            if instance is StopCommand:
                return
            for analysis in self.analyses:
                try:
                    if isinstance(instance, tuple):
                        command, *args = instance
                        result = getattr(analysis, command)(*args)
                    else:
                        result = analysis.log_likelihood_function(instance)
                    self.queue.put(result)
                except Exception as e:
                    self.queue.put(e)

    def run(self):
        """
        Run this process, completing each job in the job_queue and
        passing the result to the queue.
        """

        self.logger.debug("starting")

        self._run()

        self.logger.debug("terminating")
        self.queue.close()


class AnalysisPool:
    def __init__(self, analyses, n_cores: int):
        """
        A pool which distributes analysis evenly across
        n_cores processes.

        Callable as a log-likelihood function.

        Parameters
        ----------
        analyses
            A list of analyses to be performed in parallel
        n_cores
            The number of cores available to perform analyses over
        """
        self.analyses = analyses
        self.n_cores = n_cores

        self.n_analyses = len(analyses)
        n_processes = min(self.n_analyses, n_cores)

        analyses_per_process = math.ceil(self.n_analyses / n_processes)

        self.processes = list()

        for n in range(n_processes):
            analyses_ = analyses[
                n * analyses_per_process : (n + 1) * analyses_per_process
            ]
            process = AnalysisProcess(analyses=analyses_)

            self.processes.append(process)
            process.start()

    def __del__(self):
        """
        Called when the map goes out of scope.

        Tell each process to terminate with a StopCommand and then join
        each process with a timeout of one second.
        """
        self.terminate()

    def terminate(self):
        """
        Terminate each process and join it with a timeout of one second.
        """
        logger.debug("Deconstructing SneakyMap")

        logger.debug("Terminating processes...")
        for process in self.processes:
            process.instance_queue.put(StopCommand)

        logger.debug("Joining processes...")
        for process in self.processes:
            try:
                process.join(0.5)
            except Exception as e:
                logger.exception(e)

    def __call__(self, instance) -> float:
        """
        Evaluate the likelihood of an instance according to several
        analyses, in parallel.

        Parameters
        ----------
        instance
            Some instance

        Returns
        -------
        The total log likelihood
        """
        for process in self.processes:
            process.instance_queue.put(instance)

        return sum(self.results())

    def results(self):
        """
        Get the results from each process not necessarily in order.
        """
        count_ = 0
        results = []

        while count_ < self.n_analyses:
            for process in self.processes:
                if process.queue.empty():
                    continue
                result = process.queue.get()
                results.append(result)

                if isinstance(result, Exception):
                    raise result

                count_ += 1

        return results

    def map(
        self,
        function_name: str,
        paths: AbstractPaths,
        *args,
    ):
        """
        Call a function on each analysis in parallel.

        Parameters
        ----------
        function_name
            The name of the function to call
        paths
            The paths to the output directory
        args
            The arguments to pass to the function
        """
        for i, process in enumerate(self.processes):
            child_paths = paths.for_sub_analysis(analysis_name=f"analyses/analysis_{i}")
            process.instance_queue.put((function_name, child_paths, *args))

        self.results()



import logging
from typing import List, Optional

from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.result import Result
from .analysis import Analysis
from autofit.non_linear.samples.summary import SamplesSummary
from autofit.non_linear.samples import SamplesPDF

logger = logging.getLogger(__name__)


class CombinedResult(Result):
    def __init__(
        self,
        results: List[Result],
        samples: Optional[SamplesPDF] = None,
        samples_summary: Optional[SamplesSummary] = None,
        paths: Optional[AbstractPaths] = None,
        search_internal: Optional[object] = None,
        analysis: Optional[Analysis] = None,
    ):
        """
        A `Result` object that is composed of multiple `Result` objects. This is used to combine the results of
        multiple `Analysis` objects into a single `Result` object, for example when performing a model-fitting
        analysis where there are multiple datasets.

        Parameters
        ----------
        results
            The list of `Result` objects that are combined into this `CombinedResult` object.
        """
        super().__init__(
            samples_summary=samples_summary,
            samples=samples,
            paths=paths,
            search_internal=search_internal,
            analysis=analysis,
        )
        self.child_results = results

    def __getattr__(self, item: str):
        """
        Get an attribute of the first `Result` object in the list of `Result` objects.
        """
        if item in ("__getstate__", "__setstate__"):
            raise AttributeError(item)
        return getattr(self.child_results[0], item)

    def __iter__(self):
        return iter(self.child_results)

    def __len__(self):
        return len(self.child_results)

    def __getitem__(self, item: int) -> Result:
        """
        Get a `Result` object from the list of `Result` objects.
        """
        return self.child_results[item]



from autoconf import jax_wrapper
from autoconf.dictable import register_parser
from . import conf

conf.instance.register(__file__)

import abc
import pickle
from dill import register

from . import exc
from . import mock as m
from .non_linear.grid.grid_search import GridSearch as SearchGridSearch
from .aggregator.base import AggBase
from .database.aggregator.aggregator import GridSearchAggregator
from .graphical.expectation_propagation.history import EPHistory
from .graphical.declarative.factor.analysis import AnalysisFactor
from .graphical.declarative.collection import FactorGraphModel
from .graphical.declarative.factor.hierarchical import HierarchicalFactor
from .graphical.laplace import LaplaceOptimiser
from .non_linear.grid.grid_list import GridList
from .non_linear.samples.summary import SamplesSummary
from .non_linear.samples import SamplesMCMC
from .non_linear.samples import SamplesNest
from .non_linear.samples import Samples
from .non_linear.samples import SamplesPDF
from .non_linear.samples import Sample
from .non_linear.samples import load_from_table
from .non_linear.samples import SamplesStored
from .database.aggregator import Aggregator
from .aggregator.summary.aggregate_csv import AggregateCSV
from .aggregator.summary.aggregate_csv import ValueType
from .aggregator.summary.aggregate_images import AggregateImages
from .aggregator.summary.aggregate_fits import AggregateFITS
from .database.aggregator import Query
from autofit.aggregator.fit_interface import Fit
from .aggregator.search_output import SearchOutput
from .mapper import prior
from .mapper.model import AbstractModel
from .mapper.model import ModelInstance
from .mapper.model import ModelInstance as Instance
from .mapper.model import path_instances_of_class
from .mapper.model_mapper import ModelMapper
from .mapper.model_mapper import ModelMapper as Mapper
from .mapper.model_object import ModelObject
from .mapper.operator import DiagonalMatrix
from .mapper.prior.constant import Constant
from .mapper.prior.arithmetic.assertion import ComparisonAssertion
from .mapper.prior.arithmetic.assertion import ComparisonAssertion
from .mapper.prior.arithmetic.assertion import GreaterThanLessThanAssertion
from .mapper.prior.arithmetic.assertion import GreaterThanLessThanEqualAssertion
from .mapper.prior.deferred import DeferredArgument
from .mapper.prior.deferred import DeferredInstance
from .mapper.prior.width_modifier import AbsoluteWidthModifier
from .mapper.prior.width_modifier import RelativeWidthModifier
from .mapper.prior.width_modifier import WidthModifier
from .mapper.prior import GaussianPrior
from .mapper.prior import LogGaussianPrior
from .mapper.prior import LogUniformPrior
from .mapper.prior import TruncatedGaussianPrior
from .mapper.prior.vectorized import PriorVectorized
from .mapper.prior.abstract import Prior
from .mapper.prior.tuple_prior import TuplePrior
from .mapper.prior import UniformPrior
from .mapper.prior_model.abstract import AbstractPriorModel
from .mapper.prior_model.annotation import AnnotationPriorModel
from .mapper.prior_model.collection import Collection
from .mapper.prior_model.prior_model import Model
from .mapper.prior_model.array import Array
from .non_linear.search.abstract_search import NonLinearSearch
from .non_linear.analysis.visualize import Visualizer
from .non_linear.analysis.analysis import Analysis
from .non_linear.grid.grid_search import GridSearchResult
from .non_linear.grid.sensitivity import Sensitivity
from .non_linear.initializer import InitializerBall
from .non_linear.initializer import InitializerPrior
from .non_linear.initializer import InitializerParamBounds
from .non_linear.initializer import InitializerParamStartPoints
from .non_linear.search.mcmc.auto_correlations import AutoCorrelationsSettings
from .non_linear.search.mcmc.emcee.search import Emcee
from .non_linear.search.mcmc.zeus.search import Zeus
from .non_linear.search.nest.nautilus.search import Nautilus
from .non_linear.search.nest.dynesty.search.dynamic import DynestyDynamic
from .non_linear.search.nest.dynesty.search.static import DynestyStatic
from .non_linear.search.nest.ultranest.search import UltraNest
from .non_linear.search.mle.drawer.search import Drawer
from .non_linear.search.mle.bfgs.search import BFGS
from .non_linear.search.mle.bfgs.search import LBFGS
from .non_linear.search.mle.pyswarms.search.globe import PySwarmsGlobal
from .non_linear.search.mle.pyswarms.search.local import PySwarmsLocal
from .non_linear.paths.abstract import AbstractPaths
from .non_linear.paths import DirectoryPaths
from .non_linear.paths import DatabasePaths
from .non_linear.result import Result
from .non_linear.result import ResultsCollection
from .non_linear.settings import SettingsSearch
from .non_linear.samples.pdf import marginalize
from .text import formatter
from .text import samples_text
from .visualise import VisualiseGraph
from .interpolator import (
    LinearInterpolator,
    SplineInterpolator,
    CovarianceInterpolator,
    LinearRelationship,
)
from .tools import util

from autofit.mapper.prior.arithmetic.compound import SumPrior as Add
from autofit.mapper.prior.arithmetic.compound import MultiplePrior as Multiply
from autofit.mapper.prior.arithmetic.compound import DivisionPrior as Divide
from autofit.mapper.prior.arithmetic.compound import ModPrior as Mod
from autofit.mapper.prior.arithmetic.compound import PowerPrior as Power
from autofit.mapper.prior.arithmetic.compound import AbsolutePrior as Abs
from autofit.mapper.prior.arithmetic.compound import Log
from autofit.mapper.prior.arithmetic.compound import Log10

from . import example as ex
from . import database as db


for type_ in (
    "model",
    "collection",
    "tuple_prior",
    "dict",
    "instance",
    "Uniform",
    "LogUniform",
    "Gaussian",
    "LogGaussian",
    "TruncatedGaussian",
    "compound",
    "Constant",
):
    register_parser(type_, ModelObject.from_dict)


@register(abc.ABCMeta)
def save_abc(pickler, obj):
    pickle._Pickler.save_type(pickler, obj)


__version__ = "2026.1.21.3"


import json
from abc import abstractmethod, ABC
from pathlib import Path
from typing import Union, Optional, Dict

import numpy as np

import dill

from autoconf.dictable import from_dict


class FileOutput(ABC):
    def __new__(cls, name, path: Path):
        suffix = path.suffix
        if suffix == ".pickle":
            return super().__new__(PickleOutput)
        elif suffix == ".json":
            return super().__new__(JSONOutput)
        elif suffix == ".csv":
            return super().__new__(ArrayOutput)
        elif suffix == ".fits":
            return super().__new__(FITSOutput)
        raise ValueError(f"File {path} is not a valid output file")

    def __init__(self, name: str, path: Path):
        """
        An output file from a fit, which is either a pickle, json, csv or fits file.

        Parameters
        ----------
        name
            The name of the output file. This is the path to the file from the root output directory
            separated by '.' and without a suffix. e.g. output/directory/info.json -> directory.info
        path
            The path to the output file
        """
        self.name = name
        self.path = path

    def __str__(self):
        return f"{self.name} ({self.path})"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    @property
    @abstractmethod
    def value(self):
        pass


class ArrayOutput(FileOutput):
    @property
    def value(self) -> np.ndarray:
        """
        The array stored in the csv file
        """
        return np.loadtxt(self.path, delimiter=",")


class JSONOutput(FileOutput):
    @property
    def dict(self) -> Union[dict, list]:
        """
        The dictionary stored in the json file
        """
        with open(self.path) as f:
            return json.load(f)

    @property
    def value(self):
        """
        The object represented by the JSON
        """
        return from_dict(self.dict)

    def value_using_reference(self, reference: Optional[Dict] = None):
        """
        The object represented by the JSON
        """
        return from_dict(self.dict, reference=reference)


class PickleOutput(FileOutput):
    @property
    def value(self):
        """
        The object stored in the pickle file
        """
        with open(self.path, "rb") as f:
            return dill.load(f)


class FITSOutput(FileOutput):
    @property
    def value(self):
        """
        The contents of the fits file
        """
        from astropy.io import fits

        return fits.open(self.path)



import os
from abc import ABC
from typing import Dict, Optional, Tuple, Union

import numpy as np
import warnings

from autoconf import conf
from autofit import exc
from autofit.database.sqlalchemy_ import sa
from autofit.non_linear.fitness import Fitness
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.paths.null import NullPaths
from autofit.non_linear.search.nest.abstract_nest import AbstractNest
from autofit.non_linear.samples.sample import Sample
from autofit.non_linear.samples.nest import SamplesNest


def prior_transform(cube, model):
    phys_cube = model.vector_from_unit_vector(
        unit_vector=cube,
    )

    for i in range(len(phys_cube)):
        cube[i] = phys_cube[i]

    return cube


class AbstractDynesty(AbstractNest, ABC):
    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        iterations_per_quick_update: int = None,
        iterations_per_full_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs,
    ):
        """
        A Dynesty non-linear search.

        For a full description of Dynesty, checkout its GitHub and readthedocs webpages:

        https://github.com/joshspeagle/dynesty
        https://dynesty.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        iterations_per_full_update
            The number of iterations performed between every Dynesty back-up (via dumping the Dynesty instance as a
            pickle)
        number_of_cores
            The number of cores sampling is performed using a Python multiprocessing Pool instance.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs,
        )

        self.logger.debug(f"Creating {self.__class__.__name__} Search")

    def _fit(
        self,
        model: AbstractPriorModel,
        analysis,
    ):
        """
        Fit a model using the search and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        By default, Dynesty runs using an in-built multiprocessing Pool option. This occurs even
        if `number_of_cores=1`, because the dynesty savestate includes this pool, meaning that a resumed run can
        then increase the `number_of_cores`.

        However, certain operating systems (e.g. Windows) do not support Python multiprocessing particularly well.
        This can cause Dynesty to crash when a pool is included. If this occurs (raising a `RunTimeException`)
        a Dynesty object without a pool is created and used instead.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space.
        analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data,
            returning the log likelihood dynesty maximizes.

        Returns
        -------
        A result object comprising the Samples object that includes the maximum log likelihood instance and full
        set of accepted samples of the fit.
        """

        fitness = Fitness(
            model=model,
            analysis=analysis,
            paths=self.paths,
            fom_is_log_likelihood=True,
            resample_figure_of_merit=-1.0e99,
        )

        if not isinstance(self.paths, NullPaths):
            checkpoint_exists = os.path.exists(self.checkpoint_file)
        else:
            checkpoint_exists = False

        if checkpoint_exists:
            self.logger.info(
                "Resuming Dynesty non-linear search (previous samples found)."
            )
        else:
            self.logger.info(
                "Starting new Dynesty non-linear search (no previous samples found)."
            )

        finished = False

        while not finished:
            try:
                if (
                    conf.instance["non_linear"]["nest"][self.__class__.__name__][
                        "parallel"
                    ].get("force_x1_cpu")
                    or self.kwargs.get("force_x1_cpu")
                    or analysis._use_jax
                ):
                    raise RuntimeError

                from dynesty.pool import Pool

                with Pool(
                    njobs=self.number_of_cores,
                    loglike=fitness,
                    prior_transform=prior_transform,
                    logl_args=(model, fitness),
                    ptform_args=(model,),
                ) as pool:
                    search_internal = self.search_internal_from(
                        model=model,
                        fitness=fitness,
                        checkpoint_exists=checkpoint_exists,
                        pool=pool,
                        queue_size=self.number_of_cores,
                    )

                    finished = self.run_search_internal(search_internal=search_internal)

                    checkpoint_exists = True

            except RuntimeError:
                if not checkpoint_exists:
                    self.logger.info(
                        """
                        Your operating system does not support Python multiprocessing.

                        A single CPU non-multiprocessing Dynesty run is being performed.
                        """
                    )

                search_internal = self.search_internal_from(
                    model=model,
                    fitness=fitness,
                    checkpoint_exists=checkpoint_exists,
                    pool=None,
                    queue_size=None,
                )

                finished = self.run_search_internal(search_internal=search_internal)

                checkpoint_exists = True

            if not finished:
                self.perform_update(
                    model=model,
                    analysis=analysis,
                    search_internal=search_internal,
                    fitness=fitness,
                    during_analysis=True,
                )

        return search_internal, fitness

    def samples_info_from(self, search_internal=None):
        search_internal = search_internal or self.search_internal

        return {
            "log_evidence": np.max(search_internal.results.logz),
            "total_samples": int(np.sum(search_internal.results.ncall)),
            "total_accepted_samples": len(search_internal.results.logl),
            "time": self.timer.time if self.timer else None,
            "number_live_points": self.number_live_points,
        }

    def samples_via_internal_from(self, model, search_internal=None):
        """
        Returns a `Samples` object from the dynesty internal results.

        The samples contain all information on the parameter space sampling (e.g. the parameters,
        log likelihoods, etc.).

        The internal search results are converted from the native format used by the search to lists of values
        (e.g. `parameter_lists`, `log_likelihood_list`).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """
        search_internal = search_internal or self.search_internal

        parameter_lists = search_internal.results.samples.tolist()
        log_prior_list = model.log_prior_list_from(parameter_lists=parameter_lists)
        log_likelihood_list = list(search_internal.results.logl)

        weight_list = list(
            np.exp(
                np.asarray(search_internal.results.logwt)
                - search_internal.results.logz[-1]
            )
        )

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list,
        )

        return SamplesNest(
            model=model,
            sample_list=sample_list,
            samples_info=self.samples_info_from(search_internal=search_internal),
        )

    @property
    def search_internal(self):
        raise NotImplementedError

    def iterations_from(
        self, search_internal: "Union[NestedSampler, DynamicNestedSampler]"
    ) -> Tuple[int, int]:
        """
        Returns the next number of iterations that a dynesty call will use and the total number of iterations
        that have been performed so far.

        This is used so that the `iterations_per_full_update` input leads to on-the-fly output of dynesty results.

        It also ensures dynesty does not perform more samples than the `maxcall` input variable.

        Parameters
        ----------
        search_internal
            The Dynesty sampler (static or dynamic) which is run and performs nested sampling.

        Returns
        -------
        The next number of iterations that a dynesty run sampling will perform and the total number of iterations
        it has performed so far.
        """

        if isinstance(self.paths, NullPaths):
            maxcall = self.config_dict_run.get("maxcall")

            if maxcall is not None:
                return maxcall, maxcall
            return int(1e99), int(1e99)

        try:
            total_iterations = np.sum(search_internal.results.ncall)
        except AttributeError:
            total_iterations = 0

        if self.config_dict_run.get("maxcall") is not None:
            iterations = self.config_dict_run["maxcall"] - total_iterations

            return int(iterations), int(total_iterations)
        return self.iterations_per_full_update, int(total_iterations)

    def run_search_internal(
        self, search_internal: "Union[NestedSampler, DynamicNestedSampler]"
    ):
        """
        Run the Dynesty sampler, which could be either the static of dynamic sampler.

        The number of iterations which the sampler runs from depends on the `maxcall` input. Due to on-to-fly updates
        via `perform_update` the number of remaining iterations compared to `maxcall` is tracked between every
        `run_nested` call, and whether or not the sampler is finished is returned at the end of this function.

        A second finish criteria is used, which occurs when the dynesty run performs zero likelihood updates (because
        the sampling accrording to Dynesty's termination criteria is complete).

        Parameters
        ----------
        sampler
            The Dynesty sampler (static or dynamic) which is run and performs nested sampling.

        Returns
        -------

        """

        iterations, total_iterations = self.iterations_from(
            search_internal=search_internal
        )

        config_dict_run = {
            key: value
            for key, value in self.config_dict_run.items()
            if key != "maxcall"
        }

        if iterations > 0:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                search_internal.run_nested(
                    maxcall=iterations,
                    print_progress=not self.silence,
                    checkpoint_file=self.checkpoint_file,
                    **config_dict_run,
                )

        iterations_after_run = np.sum(search_internal.results.ncall)

        if isinstance(self.paths, NullPaths):
            return True

        return (
            total_iterations == iterations_after_run
            or total_iterations == self.config_dict_run.get("maxcall")
        )

    def write_uses_pool(self, uses_pool: bool) -> str:
        """
        If a Dynesty fit does not use a parallel pool, and is then resumed using one,
        this causes significant slow down.

        This file checks the original pool use so an exception can be raised to avoid this.

        If autofit is not outputting results to hard-disk (e.g. paths is `NullPaths`), this function is bypassed.
        """
        try:
            with open(self.paths.search_internal_path / "uses_pool.save", "w+") as f:
                if uses_pool:
                    f.write("True")
                else:
                    f.write("")
        except TypeError:
            pass

    def read_uses_pool(self) -> str:
        """
        If a Dynesty fit does not use a parallel pool, and is then resumed using one,
        this causes significant slow down.

        This file checks the original pool use so an exception can be raised to avoid this.
        """
        with open(self.paths.search_internal_path / "uses_pool.save", "r+") as f:
            return bool(f.read())

    @property
    def checkpoint_file(self) -> str:
        """
        The path to the file used for checkpointing.

        If autofit is not outputting results to hard-disk (e.g. paths is `NullPaths`), this function is bypassed.
        """
        try:
            return str(self.paths.search_internal_path / "savestate.save")
        except TypeError:
            pass

    def config_dict_test_mode_from(self, config_dict: Dict) -> Dict:
        """
        Returns a configuration dictionary for test mode meaning that the sampler terminates as quickly as possible.

        Entries which set the total number of samples of the sampler (e.g. maximum calls, maximum likelihood
        evaluations) are reduced to low values meaning it terminates nearly immediately.

        Parameters
        ----------
        config_dict
            The original configuration dictionary for this sampler which includes entries controlling how fast the
            sampler terminates.

        Returns
        -------
        A configuration dictionary where settings which control the sampler's number of samples are reduced so it
        terminates as quickly as possible.
        """
        return {
            **config_dict,
            "maxiter": 1,
            "maxcall": 1,
        }

    def live_points_init_from(self, model, fitness):
        """
        By default, dynesty live points are generated via the sampler's in-built initialization.

        However, in test-mode this would take a long time to run, thus we overwrite the initial live points
        with quickly generated samplers from the initializer.

        Parameters
        ----------
        model
        fitness

        Returns
        -------

        """

        (
            unit_parameters,
            parameters,
            log_likelihood_list,
        ) = self.initializer.samples_from_model(
            total_points=self.number_live_points,
            model=model,
            fitness=fitness,
            paths=self.paths,
            n_cores=self.number_of_cores,
        )

        init_unit_parameters = np.zeros(
            shape=(self.number_live_points, model.prior_count)
        )
        init_parameters = np.zeros(shape=(self.number_live_points, model.prior_count))
        init_log_likelihood_list = np.zeros(shape=(self.number_live_points))

        for i in range(len(parameters)):
            init_unit_parameters[i, :] = np.asarray(unit_parameters[i])
            init_parameters[i, :] = np.asarray(parameters[i])
            init_log_likelihood_list[i] = np.asarray(log_likelihood_list[i])

        live_points = [
            init_unit_parameters,
            init_parameters,
            init_log_likelihood_list,
        ]

        blobs = np.asarray(self.number_live_points * [False])

        live_points.append(blobs)

        return live_points

    def search_internal_from(
        self,
        model: AbstractPriorModel,
        fitness,
        checkpoint_exists: bool,
        pool: Optional,
        queue_size: Optional[int],
    ):
        raise NotImplementedError()

    def output_search_internal(self, search_internal):

        self.paths.save_search_internal(
            obj=search_internal,
        )

        try:
            os.remove(self.checkpoint_file)
        except (TypeError, FileNotFoundError):
            pass

    def check_pool(self, uses_pool: bool, pool):
        if (uses_pool and pool is None) or (not uses_pool and pool is not None):
            raise exc.SearchException(
                """
                A Dynesty sampler has been loaded and its pool type is not the same as the input pool type.

                This means that the original samples in dynesty were computed with or without a 
                multiprocessing pool, whereas the run is now trying to use a multiprocessing pool.

                This could indiciate the number of cores have change values or Python multiprocessing
                has been disabled and then enabled.
                """
            )

    @property
    def number_live_points(self):
        raise NotImplementedError()



import configparser
import logging
import os
import random
from abc import ABC, abstractmethod
from typing import Dict, Tuple, List, Optional

import numpy as np

from autofit import exc
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.parallel import SneakyPool

logger = logging.getLogger(__name__)


class AbstractInitializer(ABC):

    @abstractmethod
    def _generate_unit_parameter_list(self, model):
        pass

    def info_from_model(self, model : AbstractPriorModel) -> str:
        raise NotImplementedError

    @staticmethod
    def figure_of_metric(args) -> Optional[float]:
        fitness, parameter_list = args
        try:
            figure_of_merit = fitness(parameters=parameter_list)

            if np.isnan(figure_of_merit) or figure_of_merit < -1e98:
                return None

            return figure_of_merit
        except exc.FitException:
            return None

    def samples_from_model(
            self,
            total_points: int,
            model: AbstractPriorModel,
            fitness,
            paths: AbstractPaths,
            use_prior_medians: bool = False,
            test_mode_samples: bool = True,
            n_cores: int = 1,
    ):
        """
        Generate the initial points of the non-linear search, by randomly drawing unit values from a uniform
        distribution between the ball_lower_limit and ball_upper_limit values.

        Parameters
        ----------
        total_points
            The number of points in non-linear paramemter space which initial points are created for.
        model
            An object that represents possible instances of some model with a given dimensionality which is the number
            of free dimensions of the model.
        """

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1" and test_mode_samples:
            return self.samples_in_test_mode(total_points=total_points, model=model)

        if n_cores == 1:
            return self.samples_jax(
                total_points=total_points,
                model=model,
                fitness=fitness,
                use_prior_medians=use_prior_medians
            )

        unit_parameter_lists = []
        parameter_lists = []
        figures_of_merit_list = []

        sneaky_pool = SneakyPool(n_cores, fitness, paths)

        logger.info(f"Generating initial samples of model using {n_cores} cores")

        while len(figures_of_merit_list) < total_points:
            remaining_points = total_points - len(figures_of_merit_list)
            batch_size = min(remaining_points, n_cores)
            parameter_lists_ = []
            unit_parameter_lists_ = []

            for _ in range(batch_size):
                if not use_prior_medians:
                    unit_parameter_list = self._generate_unit_parameter_list(model)
                else:
                    unit_parameter_list = [0.5] * model.prior_count

                parameter_list = model.vector_from_unit_vector(
                    unit_vector=unit_parameter_list
                )

                parameter_lists_.append(parameter_list)
                unit_parameter_lists_.append(unit_parameter_list)

            for figure_of_merit, unit_parameter_list, parameter_list in zip(
                    sneaky_pool.map(
                        function=self.figure_of_metric,
                        args_list=[(fitness, parameter_list) for parameter_list in parameter_lists_],
                        log_info=False
                    ),
                    unit_parameter_lists_,
                    parameter_lists_,
            ):
                if figure_of_merit is not None:
                    unit_parameter_lists.append(unit_parameter_list)
                    parameter_lists.append(parameter_list)
                    figures_of_merit_list.append(figure_of_merit)

        if total_points > 1 and np.allclose(
                a=figures_of_merit_list[0], b=figures_of_merit_list[1:]
        ):
            raise exc.InitializerException(
                """
                The initial samples all have the same figure of merit (e.g. log likelihood values).

                The non-linear search will therefore not progress correctly.

                Possible causes for this behaviour are:

                - The `log_likelihood_function` of the analysis class is defined incorrectly.
                - The model parameterization creates numerically inaccurate log likelihoods.
                - The`log_likelihood_function`  is always returning `nan` values.            
                """
            )

        logger.info(f"Initial samples generated, starting non-linear search")

        return unit_parameter_lists, parameter_lists, figures_of_merit_list

    def samples_jax(
            self,
            total_points: int,
            model: AbstractPriorModel,
            fitness,
            use_prior_medians: bool = False,
    ):
        """
        Generate the initial points of the non-linear search, by randomly drawing unit values from a uniform
        distribution between the ball_lower_limit and ball_upper_limit values.

        Parameters
        ----------
        total_points
            The number of points in non-linear paramemter space which initial points are created for.
        model
            An object that represents possible instances of some model with a given dimensionality which is the number
            of free dimensions of the model.
        """

        unit_parameter_lists = []
        parameter_lists = []
        figures_of_merit_list = []

        logger.info(f"Generating initial samples of model using JAX LH Function cores")

        while len(figures_of_merit_list) < total_points:

            if not use_prior_medians:
                unit_parameter_list = self._generate_unit_parameter_list(model)
            else:
                unit_parameter_list = [0.5] * model.prior_count

            parameter_list = model.vector_from_unit_vector(
                unit_vector=unit_parameter_list
            )

            figure_of_merit = self.figure_of_metric((fitness, parameter_list))

            if figure_of_merit is not None:
                unit_parameter_lists.append(unit_parameter_list)
                parameter_lists.append(parameter_list)
                figures_of_merit_list.append(figure_of_merit)

        if total_points > 1 and np.allclose(
                a=figures_of_merit_list[0], b=figures_of_merit_list[1:]
        ):
            raise exc.InitializerException(
                """
                The initial samples all have the same figure of merit (e.g. log likelihood values).

                The non-linear search will therefore not progress correctly.

                Possible causes for this behaviour are:

                - The `log_likelihood_function` of the analysis class is defined incorrectly.
                - The model parameterization creates numerically inaccurate log likelihoods.
                - The`log_likelihood_function`  is always returning `nan` values.            
                """
            )

        logger.info(f"Initial samples generated, starting non-linear search")

        return unit_parameter_lists, parameter_lists, figures_of_merit_list

    def samples_in_test_mode(self, total_points: int, model: AbstractPriorModel):
        """
        Generate the initial points of the non-linear search in test mode. Like normal, test model draws points, by
        randomly drawing unit values from a uniform distribution between the ball_lower_limit and ball_upper_limit
        values.

        However, the log likelihood function is bypassed and all likelihoods are returned with a value -1.0e99. This
        is so that integration testing of large-scale model-fitting projects can be performed efficiently by bypassing
        sampling of points using the `log_likelihood_function`.

        Parameters
        ----------
        total_points
            The number of points in non-linear paramemter space which initial points are created for.
        model
            An object that represents possible instances of some model with a given dimensionality which is the number
            of free dimensions of the model.
        """

        logger.warning(
            f"TEST MODE ON: SAMPLES BEING ASSIGNED ABRITRARY LARGE LIKELIHOODS"
        )

        unit_parameter_lists = []
        parameter_lists = []
        figure_of_merit_list = []

        point_index = 0

        figure_of_merit = -1.0e99

        while point_index < total_points:
            try:
                unit_parameter_list = self._generate_unit_parameter_list(model)
                parameter_list = model.vector_from_unit_vector(
                    unit_vector=unit_parameter_list
                )
                model.instance_from_vector(vector=parameter_list)
                unit_parameter_lists.append(unit_parameter_list)
                parameter_lists.append(parameter_list)
                figure_of_merit_list.append(figure_of_merit)
                figure_of_merit *= 10.0
                point_index += 1
            except exc.FitException:
                pass

        return unit_parameter_lists, parameter_lists, figure_of_merit_list


class InitializerParamBounds(AbstractInitializer):
    def __init__(
        self,
        parameter_dict: Dict[Prior, Tuple[float, float]],
        lower_limit=0.0,
        upper_limit=1.0,
    ):
        """
        Initializer which uses the bounds on input parameters as the starting point for the search (e.g. where
        an MLE optimization starts or MCMC walkers are initialized).

        Parameters
        ----------
        parameter_dict
            A dictionary mapping each parameter path to bounded ranges of physical values that
            are where the search begins.
        lower_limit
            A default, unit lower limit used when a prior is not specified
        upper_limit
            A default, unit upper limit used when a prior is not specified
        """
        
        self.parameter_dict = parameter_dict
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

        self._generated_warnings = set()

    def _generate_unit_parameter_list(self, model: AbstractPriorModel) -> List[float]:
        """
        Generate a unit vector for the model. The default limits are used for any
        priors which the model has but are not found in the parameter dict.

        Parameters
        ----------
        model
            A model for which initial points are required

        Returns
        -------
        A unit vector
        """

        unit_parameter_list = []
        for prior in model.priors_ordered_by_id:

            try:
                lower, upper = map(prior.unit_value_for, self.parameter_dict[prior])
                value = random.uniform(lower, upper)
            except KeyError:
                key = ".".join(model.path_for_prior(prior))
                if key not in self._generated_warnings:
                    logger.warning(
                        f"Range for {key} not set in the InitializerParamBounds. "
                        f"Using defaults."
                    )
                    self._generated_warnings.add(key)

                lower = self.lower_limit
                upper = self.upper_limit

                value = prior.unit_value_for(prior.random(lower, upper))

            unit_parameter_list.append(value)

        return unit_parameter_list

    def info_from_model(self, model : AbstractPriorModel) -> str:
        """
        Returns a string showing the bounds of the parameters in the initializer.
        """
        info = "Total Free Parameters = " + str(model.prior_count) + "\n"
        info += "Total Starting Points = " + str(len(self.parameter_dict)) + "\n\n"
        for prior in model.priors_ordered_by_id:

            key = ".".join(model.path_for_prior(prior))

            try:

                value = self.info_value_from(self.parameter_dict[prior])

                info += f"{key}: Start[{value}]\n"

            except KeyError:

                info += f"{key}: {prior})\n"

        return info

    def info_value_from(self, value : Tuple[float, float]) -> Tuple[float, float]:
        """
        Returns the value that is used to display the bounds of the parameters in the initializer.

        This function simply returns the input value, but it can be overridden in subclasses for diffferent
        initializers.

        Parameters
        ----------
        value
            The value to be displayed in the initializer info which is a tuple of the lower and upper bounds of the
            parameter.
        """
        return value


class InitializerParamStartPoints(InitializerParamBounds):
    def __init__(
        self,
        parameter_dict: Dict[Prior, float],
    ):
        """
        Initializer which input values of the parameters as the starting point for the search (e.g. where
        an MLE optimization starts or MCMC walkers are initialized).

        Parameters
        ----------
        parameter_dict
            A dictionary mapping each parameter path to the starting point physical values that
            are where the search begins.
        lower_limit
            A default, unit lower limit used when a prior is not specified
        upper_limit
            A default, unit upper limit used when a prior is not specified
        """
        parameter_dict_new = {}

        for key, value in parameter_dict.items():
            parameter_dict_new[key] = (value - 1.0e-8, value + 1.0e-8)

        super().__init__(parameter_dict=parameter_dict_new)

    def info_value_from(self, value : Tuple[float, float]) -> float:
        """
        Returns the value that is used to display the starting point of the parameters in the initializer.

        This function returns the mean of the input value, as the starting point is a single value in the center of the
        bounds.

        Parameters
        ----------
        value
            The value to be displayed in the initializer info which is a tuple of the lower and upper bounds of the
            parameter.
        """
        return (value[1] + value[0]) / 2.0


class Initializer(AbstractInitializer):
    def __init__(self, lower_limit: float, upper_limit: float):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        `NonLinearSearch` to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.
        """
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    @classmethod
    def from_config(cls, config):
        """
        Load the Initializer from a non_linear config file.
        """

        try:
            initializer = config("initialize", "method")

        except configparser.NoSectionError:
            return None

        if initializer in "prior":
            return InitializerPrior()

        elif initializer in "ball":
            ball_lower_limit = config("initialize", "ball_lower_limit")
            ball_upper_limit = config("initialize", "ball_upper_limit")

            return InitializerBall(
                lower_limit=ball_lower_limit, upper_limit=ball_upper_limit
            )

    def _generate_unit_parameter_list(self, model):
        return model.random_unit_vector_within_limits(
            lower_limit=self.lower_limit, upper_limit=self.upper_limit
        )


class InitializerPrior(Initializer):
    def __init__(self):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        `NonLinearSearch` to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.

        The InitializerPrior class generates from the priors, by drawing all values as unit values between 0.0 and 1.0
        and mapping them to physical values via the prior.
        """
        super().__init__(lower_limit=0.0, upper_limit=1.0)


class InitializerBall(Initializer):
    def __init__(self, lower_limit: float, upper_limit: float):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        `NonLinearSearch` to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.

        The InitializerBall class generates the samples in a small compact volume or 'ball' in parameter space, which is
        the recommended initialization strategy for the MCMC `NonLinearSearch` Emcee.

        Parameters
        ----------
        lower_limit
            The lower limit of the uniform distribution unit values are drawn from when initializing walkers in a small
            compact ball.
        upper_limit
            The upper limit of the uniform distribution unit values are drawn from when initializing walkers in a small
            compact ball.
        """
        super().__init__(lower_limit=lower_limit, upper_limit=upper_limit)



import inspect
from typing import Type, Dict
import typing

from autofit.mapper.prior_model.attribute_pair import AttributeNameValue


class PriorModelNameValue(AttributeNameValue):
    @property
    def prior_model(self):
        return self.value


def gather_namespaces(cls: Type) -> Dict[str, Dict]:
    """
    Recursively gather the globals and locals for a given class and its parent classes.
    """
    namespaces = {}

    try:
        for base in inspect.getmro(cls):
            if base is object:
                continue

            module = inspect.getmodule(base)
            if module:
                namespaces.update(vars(module))
    except AttributeError:
        pass

    namespaces.update(vars(typing))

    return namespaces



import math
import numpy as np
from typing import Dict, List, Optional
import warnings

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.search.mcmc.auto_correlations import AutoCorrelations
from autofit.non_linear.search.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.samples.pdf import SamplesPDF
from autofit.non_linear.samples.samples import Sample

from autofit.non_linear.samples.samples import to_instance

from autofit import exc


class SamplesMCMC(SamplesPDF):
    def __init__(
        self,
        model: AbstractPriorModel,
        sample_list: List[Sample],
        samples_info: Optional[Dict] = None,
        auto_correlation_settings: Optional[AutoCorrelationsSettings] = None,
        auto_correlations: Optional[AutoCorrelations] = None,
    ):
        """
        Contains the samples of the non-linear search, including parameter values, log likelihoods,
        weights and other quantites.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        Attributes
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        auto_correlation_settings
            Customizes and performs auto correlation calculations performed during and after the search.
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        """

        self.auto_correlations = auto_correlations
        self.auto_correlation_settings = auto_correlation_settings

        super().__init__(
            model=model,
            sample_list=sample_list,
            samples_info=samples_info,
        )

    def __add__(self, other: "SamplesMCMC") -> "SamplesMCMC":
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        Parameters
        ----------
        other
            Another Samples class

        Returns
        -------
        A class that combined the samples of the two Samples objects.
        """

        self._check_addition(other=other)

        warnings.warn(
            f"Addition of {self.__class__.__name__} cannot retain results in native format. "
            "Visualization of summed samples diabled.",
            exc.SamplesWarning,
        )

        return self.__class__(
            model=self.model,
            sample_list=self.sample_list + other.sample_list,
            samples_info=self.samples_info,
            auto_correlation_settings=self.auto_correlation_settings,
        )

    @classmethod
    def from_list_info_and_model(
        cls,
        sample_list,
        samples_info,
        model: AbstractPriorModel,
    ):
        try:
            auto_correlation_settings = AutoCorrelationsSettings(
                check_for_convergence=True,
                check_size=samples_info["check_size"],
                required_length=samples_info["required_length"],
                change_threshold=samples_info["change_threshold"],
            )
        except (KeyError, NameError):
            auto_correlation_settings = None

        return cls(
            model=model,
            sample_list=sample_list,
            samples_info=samples_info,
            auto_correlation_settings=auto_correlation_settings,
        )

    @property
    def pdf_converged(self):
        """
        To analyse and visualize samples, the analysis must be sufficiently converged to produce smooth enough PDF
        for analysis.

        This property checks whether the non-linear search's samples are sufficiently converged for analysis and
        visualization.

        Emcee samples can be analysed irrespective of how long the sampler has run, albeit low run times will likely
        produce inaccurate results.
        """
        try:
            if len(self.parameter_lists) == 0:
                return False
            return True
        except ValueError:
            return False

    @property
    def converged(self) -> bool:
        """
        Whether the emcee samples have converged on a solution or if they are still in a burn-in period, based on the
        auto correlation times of parameters.
        """
        return self.auto_correlations.check_if_converged(
            total_samples=self.total_samples
        )

    @to_instance
    def median_pdf(self, as_instance: bool = True) -> [float]:
        """
        The median of the probability density function (PDF) of every parameter marginalized in 1D, returned
        as a list of values.

        This is computed by binning all sampls after burn-in into a histogram and take its median (e.g. 50%) value.
        """

        if self.pdf_converged:
            return [
                float(np.percentile(self.parameters_extract[i, :], [50]))
                for i in range(self.model.prior_count)
            ]

        return self.max_log_likelihood(as_instance=False)

    @to_instance
    def values_at_sigma(self, sigma: float) -> [float]:
        """
        The value of every parameter marginalized in 1D at an input sigma value of its probability density function
        (PDF), returned as two lists of values corresponding to the lower and upper values parameter values.

        For example, if sigma is 1.0, the marginalized values of every parameter at 31.7% and 68.2% percentiles of each
        PDF is returned.

        This does not account for covariance between parameters. For example, if two parameters (x, y) are degenerate
        whereby x decreases as y gets larger to give the same PDF, this function will still return both at their
        upper values. Thus, caution is advised when using the function to reperform a model-fits.

        For Emcee, if the samples have converged this is estimated by binning the samples after burn-in into a
        histogram and taking the parameter values at the input PDF %.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        limit = math.erf(0.5 * sigma * math.sqrt(2))

        if self.pdf_converged:
            samples = self.parameters_extract

            return [
                tuple(
                    np.percentile(samples[i, :], [100.0 * (1.0 - limit), 100.0 * limit])
                )
                for i in range(self.model.prior_count)
            ]

        parameters_min = list(
            np.min(self.parameter_lists[-self.unconverged_sample_size :], axis=0)
        )
        parameters_max = list(
            np.max(self.parameter_lists[-self.unconverged_sample_size :], axis=0)
        )

        return [
            (parameters_min[index], parameters_max[index])
            for index in range(len(parameters_min))
        ]

    @property
    def total_steps(self) -> int:
        return self.samples_info["total_steps"]

    @property
    def total_walkers(self) -> int:
        return self.samples_info["total_walkers"]

    @property
    def log_evidence(self):
        return None



from .model import Object
from ..sqlalchemy_ import sa


class NoneInstance(Object):
    __tablename__ = "none"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "none"}

    def __call__(self):
        return None


class Collection(Object):
    """
    A tuple or list
    """

    __tablename__ = "collection"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "collection"}

    @classmethod
    def _from_object(cls, source):
        instance = cls()
        instance.cls = type(source)
        instance._add_children([(str(i), item) for i, item in enumerate(source)])
        return instance

    def __call__(self):
        return self.cls(
            [
                child()
                for child in sorted(self.children, key=lambda child: int(child.name))
            ]
        )


class Instance(Object):
    """
    An instance, such as a class instance
    """

    __tablename__ = "instance"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "instance"}

    @classmethod
    def _from_object(cls, source):
        instance = cls()
        instance.cls = type(source)
        if hasattr(source, "__getstate__"):
            instance._add_children(source.__getstate__().items())
        elif hasattr(source, "__dict__"):
            instance._add_children(source.__dict__.items())
        return instance


class Value(Object):
    """
    A float
    """

    __tablename__ = "value"

    __mapper_args__ = {"polymorphic_identity": "value"}

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    value = sa.Column(sa.Float)

    @classmethod
    def _from_object(cls, source):
        instance = cls()
        instance.value = source
        return instance

    def __call__(self):
        return self.value


class StringValue(Object):
    """
    A string
    """

    __tablename__ = "string_value"

    __mapper_args__ = {"polymorphic_identity": "string_value"}

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    value = sa.Column(sa.String)

    @classmethod
    def _from_object(cls, source):
        instance = cls()
        instance.value = source
        return instance

    def __call__(self):
        return self.value



from functools import wraps
import numpy as np
import warnings

from autoconf import conf

from autofit.non_linear.plot import SamplesPlotter
from autofit.non_linear.plot.samples_plotters import skip_plot_in_test_mode


def log_value_error(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        Prevent an exception terminating the run if visualization fails due to convergence not yet being reached.

        Searches attempt to perform visualization every `iterations_per_full_update`, however these visualization calls
        may occur before the search has converged on enough of parameter to successfully perform visualization.

        This can lead the search to raise an exception which terminates the Python script, when we instead
        want the code to continue running, to continue the search and perform visualization on a subsequent iteration
        once convergence has been achieved.

        This wrapper catches these exceptions, logs them so the user can see visualization failed and then
        continues the code without raising an exception in a way that terminates the script.

        This wrapper is specific to Dynesty, which raises a `ValueError` when visualization is performed before
        convergence has been achieved.

        Parameters
        ----------
        self
            An instance of a `SearchPlotter` class.
        args
            The arguments used to perform a visualization of the search.
        kwargs
            The keyword arguments used to perform a visualization of the search.
        """
        try:
            return func(self, *args, **kwargs)
        except (ValueError, KeyError, AssertionError, IndexError, TypeError, RuntimeError, np.linalg.LinAlgError):
            pass

    return wrapper


class NestPlotter(SamplesPlotter):

    @skip_plot_in_test_mode
    @log_value_error
    def corner_anesthetic(self, **kwargs):
        """
        Plots a corner plot via the visualization library `anesthetic`.

        This plots a corner plot including the 1-D and 2-D marginalized posteriors.
        """

        config_dict = conf.instance["visualize"]["plots_settings"]["corner_anesthetic"]

        from anesthetic.samples import NestedSamples
        from anesthetic import make_2d_axes
        import matplotlib.pylab as pylab

        params = {'font.size' : int(config_dict["fontsize"])}
        pylab.rcParams.update(params)

        figsize = (
            self.model.total_free_parameters * config_dict["figsize_per_parammeter"],
            self.model.total_free_parameters * config_dict["figsize_per_parammeter"]
        )

        samples = NestedSamples(
            np.asarray(self.samples.parameter_lists),
            weights=self.samples.weight_list,
            columns=self.model.parameter_labels_with_superscripts_latex
        )

        from pandas.errors import SettingWithCopyWarning
        warnings.filterwarnings("ignore", category=SettingWithCopyWarning)

        fig, axes = make_2d_axes(
            self.model.parameter_labels_with_superscripts_latex,
            figsize=figsize,
            facecolor=config_dict["facecolor"],
        )

        warnings.filterwarnings("default", category=SettingWithCopyWarning)

        # prior = samples.prior()
        # prior.plot_2d(axes, alpha=0.9, label="prior")
        samples.plot_2d(
            axes,
            alpha=config_dict["alpha"],
            label="posterior",
        )
        axes.iloc[-1, 0].legend(
            bbox_to_anchor=(len(axes) / 2, len(axes)),
            loc='lower center',
            ncols=2
        )

        self.output.to_figure(auto_filename="corner_anesthetic")
        self.close()



import os
from typing import Dict, Optional

import numpy as np

from autoconf import conf

from autofit.database.sqlalchemy_ import sa
from autofit.mapper.model_mapper import ModelMapper
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.fitness import Fitness
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.search.mcmc.abstract_mcmc import AbstractMCMC
from autofit.non_linear.search.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.search.mcmc.auto_correlations import AutoCorrelations
from autofit.non_linear.samples.sample import Sample
from autofit.non_linear.samples.mcmc import SamplesMCMC


class Emcee(AbstractMCMC):
    __identifier_fields__ = ("nwalkers",)

    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        initializer: Optional[Initializer] = None,
        auto_correlation_settings=AutoCorrelationsSettings(),
        iterations_per_quick_update: int = None,
        iterations_per_full_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs,
    ):
        """
        An Emcee non-linear search.

        For a full description of Emcee, checkout its Github and readthedocs webpages:

        https://github.com/dfm/emcee

        https://emcee.readthedocs.io/en/stable/

        If you use `Emcee` as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        auto_correlation_settings
            Customizes and performs auto correlation calculations performed during and after the search.
        number_of_cores
            The number of cores sampling is performed using a Python multiprocessing Pool instance.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            initializer=initializer,
            auto_correlation_settings=auto_correlation_settings,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs,
        )

        self.logger.debug("Creating Emcee Search")

        # TODO : Emcee visualization tools rely on the .hdf file and thus require that the search internal is
        # TODO : On hard-disk, which this forces to occur.

        conf.instance["output"]["search_internal"] = True

    def _fit(self, model: AbstractPriorModel, analysis):
        """
        Fit a model using Emcee and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """
        import emcee

        fitness = Fitness(
            model=model,
            analysis=analysis,
            paths=self.paths,
            fom_is_log_likelihood=False,
            resample_figure_of_merit=-np.inf,
        )

        pool = self.make_sneaky_pool(fitness)

        try:
            backend = emcee.backends.HDFBackend(filename=self.backend_filename)
        except TypeError:
            backend = None

        search_internal = emcee.EnsembleSampler(
            nwalkers=self.config_dict_search["nwalkers"],
            ndim=model.prior_count,
            log_prob_fn=fitness.call_wrap,
            backend=backend,
            pool=pool,
        )

        try:
            state = search_internal.get_last_sample()
            samples = self.samples_from(model=model, search_internal=search_internal)

            total_iterations = search_internal.iteration

            if samples.converged:
                iterations_remaining = 0
            else:
                iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

                self.logger.info(
                    "Resuming Emcee non-linear search (previous samples found)."
                )

        except AttributeError:
            (
                unit_parameter_lists,
                parameter_lists,
                log_posterior_list,
            ) = self.initializer.samples_from_model(
                total_points=search_internal.nwalkers,
                model=model,
                fitness=fitness,
                paths=self.paths,
                n_cores=self.number_of_cores,
            )

            self.plot_start_point(
                parameter_vector=parameter_lists[0],
                model=model,
                analysis=analysis,
            )

            state = np.zeros(shape=(search_internal.nwalkers, model.prior_count))

            self.logger.info(
                "Starting new Emcee non-linear search (no previous samples found)."
            )

            for index, parameters in enumerate(parameter_lists):
                state[index, :] = np.asarray(parameters)

            total_iterations = 0
            iterations_remaining = self.config_dict_run["nsteps"]

        while iterations_remaining > 0:
            if self.iterations_per_full_update > iterations_remaining:
                iterations = iterations_remaining
            else:
                iterations = self.iterations_per_full_update

            for sample in search_internal.sample(
                initial_state=state,
                iterations=iterations,
                progress=True,
                skip_initial_state_check=True,
                store=True,
            ):
                pass

            state = search_internal.get_last_sample()

            total_iterations += iterations
            iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

            samples = self.samples_from(model=model, search_internal=search_internal)

            if self.auto_correlation_settings.check_for_convergence:
                if (
                    search_internal.iteration
                    > self.auto_correlation_settings.check_size
                ):
                    if samples.converged:
                        iterations_remaining = 0

            if iterations_remaining > 0:
                self.perform_update(
                    model=model,
                    analysis=analysis,
                    search_internal=search_internal,
                    fitness=fitness,
                    during_analysis=True,
                )

        return search_internal, fitness

    def output_search_internal(self, search_internal):
        """
        Output the sampler results to hard-disk in their internal format.

        Emcee uses a backend to store and load results, therefore the outputting of the search internal to a
        dill file is disabled.

        Parameters
        ----------
        sampler
            The nautilus sampler object containing the results of the model-fit.
        """
        pass

    def samples_info_from(self, search_internal=None):
        search_internal = search_internal or self.backend

        auto_correlations = self.auto_correlations_from(search_internal=search_internal)

        return {
            "check_size": auto_correlations.check_size,
            "required_length": auto_correlations.required_length,
            "change_threshold": auto_correlations.change_threshold,
            "total_walkers": len(search_internal.get_chain()[0, :, 0]),
            "total_steps": len(search_internal.get_log_prob()),
            "time": self.timer.time if self.timer else None,
        }

    def samples_via_internal_from(self, model, search_internal=None):
        """
        Returns a `Samples` object from the emcee internal results.

        The samples contain all information on the parameter space sampling (e.g. the parameters,
        log likelihoods, etc.).

        The internal search results are converted from the native format used by the search to lists of values
        (e.g. `parameter_lists`, `log_likelihood_list`).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """

        search_internal = search_internal or self.backend

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            samples_after_burn_in = search_internal.get_chain(
                discard=5, thin=5, flat=True
            )

        else:
            auto_correlations = self.auto_correlations_from(
                search_internal=search_internal
            )

            discard = int(3.0 * np.max(auto_correlations.times))
            thin = int(np.max(auto_correlations.times) / 2.0)
            samples_after_burn_in = search_internal.get_chain(
                discard=discard, thin=thin, flat=True
            )

        parameter_lists = samples_after_burn_in.tolist()

        log_prior_list = model.log_prior_list_from(parameter_lists=parameter_lists)

        total_samples = len(parameter_lists)

        log_posterior_list = search_internal.get_log_prob(flat=True)[
            -total_samples - 1 : -1
        ].tolist()

        log_likelihood_list = [
            log_posterior - log_prior
            for log_posterior, log_prior in zip(log_posterior_list, log_prior_list)
        ]

        weight_list = len(log_likelihood_list) * [1.0]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list,
        )

        return SamplesMCMC(
            model=model,
            sample_list=sample_list,
            samples_info=self.samples_info_from(search_internal=search_internal),
            auto_correlation_settings=self.auto_correlation_settings,
            auto_correlations=self.auto_correlations_from(
                search_internal=search_internal
            ),
        )

    def auto_correlations_from(self, search_internal=None):
        import emcee

        search_internal = search_internal or self.backend

        times = search_internal.get_autocorr_time(tol=0)

        previous_auto_correlation_times = emcee.autocorr.integrated_time(
            x=search_internal.get_chain()[
                : -self.auto_correlation_settings.check_size, :, :
            ],
            tol=0,
        )

        return AutoCorrelations(
            check_size=self.auto_correlation_settings.check_size,
            required_length=self.auto_correlation_settings.required_length,
            change_threshold=self.auto_correlation_settings.change_threshold,
            times=times,
            previous_times=previous_auto_correlation_times,
        )

    def config_dict_test_mode_from(self, config_dict: Dict) -> Dict:
        """
        Returns a configuration dictionary for test mode meaning that the sampler terminates as quickly as possible.

        Entries which set the total number of samples of the sampler (e.g. maximum calls, maximum likelihood
        evaluations) are reduced to low values meaning it terminates nearly immediately.

        Parameters
        ----------
        config_dict
            The original configuration dictionary for this sampler which includes entries controlling how fast the
            sampler terminates.

        Returns
        -------
        A configuration dictionary where settings which control the sampler's number of samples are reduced so it
        terminates as quickly as possible.
        """
        return {
            **config_dict,
            "nwalkers": 20,
            "nsteps": 10,
        }

    @property
    def backend_filename(self):
        return self.paths.search_internal_path / "search_internal.hdf"

    @property
    def backend(self) -> "emcee.backends.HDFBackend":
        """
        The `Emcee` hdf5 backend, which provides access to all samples, likelihoods, etc. of the non-linear search.

        The sampler is described in the "Results" section at https://dynesty.readthedocs.io/en/latest/quickstart.html
        """
        import emcee

        if os.path.isfile(self.backend_filename):
            return emcee.backends.HDFBackend(filename=str(self.backend_filename))
        else:
            raise FileNotFoundError(
                f"The file search_internal.hdf does not exist at the path {self.paths.search_internal_path}"
            )



import inspect
import json
import os
import sys
import zipfile
from contextlib import contextmanager
from functools import wraps
from pathlib import Path
from typing import Union

import numpy as np

from autoconf import conf


def split_paths(func):
    """
    Split string paths if they are passed.

    e.g. "lens.mass.centre" -> ["lens", "mass", "centre"]
    """

    @wraps(func)
    def wrapper(self, paths):
        paths = [path.split(".") if isinstance(path, str) else path for path in paths]
        return func(self, paths)

    return wrapper


class IntervalCounter:
    def __init__(self, interval):
        self.count = 0
        self.interval = interval

    def __call__(self):
        if self.interval == -1:
            return False
        self.count += 1
        return self.count % self.interval == 0


def zip_directory(source_directory, output=None):
    output = output or f"{source_directory}.zip"
    with zipfile.ZipFile(output, "w", zipfile.ZIP_DEFLATED) as f:
        for root, dirs, files in os.walk(source_directory):
            for file in files:
                f.write(
                    os.path.join(root, file),
                    os.path.join(root[len(str(source_directory)) :], file),
                )


def open_(filename, *flags):
    directory = Path(filename)
    os.makedirs(directory.parent, exist_ok=True)
    return open(filename, *flags)


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def numpy_array_to_json(
    array: np.ndarray, file_path: Union[Path, str], overwrite: bool = False
):
    """
    Write a NumPy array to a json file.

    Parameters
    ----------
    array
        The array that is written to json.
    file_path
        The full path of the file that is output, including the file name and `.json` extension.
    overwrite
        If `True` and a file already exists with the input file_path the .json file is overwritten. If
        `False`, an error will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_json(array_2d=array_2d, file_path='/path/to/file/filename.json', overwrite=True)
    """

    file_dir = os.path.split(file_path)[0]

    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    with open(file_path, "w+") as f:
        json.dump(array.tolist(), f)


def numpy_array_from_json(file_path: Union[Path, str]):
    """
    Read a 1D NumPy array from a .json file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .json files loaded in DS9.

    Parameters
    ----------
    file_path
        The full path of the file that is loaded, including the file name and ``.json`` extension.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .json file.

    Examples
    --------
    array_2d = numpy_array_from_json(file_path='/path/to/file/filename.json')
    """
    with open(file_path, "r") as f:
        return np.asarray(json.load(f))


def info_whitespace():
    return conf.instance["general"]["output"]["info_whitespace_length"]



from autofit.database.query import condition as c

from .abstract import AbstractQuery


class AttributeQuery(AbstractQuery):
    @property
    def fit_query(self) -> str:
        """
        The SQL string produced by this query. This is applied directly to the database.
        """
        return f"SELECT id FROM fit WHERE {self.condition}"


class Attribute:
    def __init__(self, attribute: str):
        """
        Some direct attribute of the Fit class

        Parameters
        ----------
        attribute
            The name of that attribute
        """
        self.attribute = attribute

    def _make_query(self, cls, value) -> AttributeQuery:
        """
        Create a query against this attribute

        Parameters
        ----------
        cls
            An AttributeCondition that describes the query
        value
            The value that the attribute is compared to

        Returns
        -------
        A query on ids of the fit table
        """
        return AttributeQuery(cls(attribute=self.attribute, value=value))

    def __eq__(self, other) -> AttributeQuery:
        """
        Check whether an attribute, such as a search name, is equal
        to some value
        """
        return self._make_query(cls=c.EqualityAttributeCondition, value=other)

    def in_(self, item: str) -> AttributeQuery:
        """
        Check whether an attribute is contained within a substring
        """
        return self._make_query(cls=c.InAttributeCondition, value=item)

    def contains(self, item: str) -> AttributeQuery:
        """
        Check whether an attribute, such as a search name, contains
        some string
        """
        return self._make_query(cls=c.ContainsAttributeCondition, value=item)


class BooleanAttribute(Attribute, AttributeQuery):
    def __init__(self, attribute):
        super().__init__(attribute)
        super(AttributeQuery, self).__init__(c.AttributeCondition(attribute))

    def __hash__(self):
        return hash(str(self))


class ChildQuery(AttributeQuery):
    def __init__(self, predicate: AbstractQuery):
        super().__init__(predicate)

    @property
    def condition(self):
        return f"parent_id in ({super().condition})"


class BestFitQuery(ChildQuery):
    def __init__(self, predicate: AbstractQuery):
        super().__init__(predicate)

    @property
    def fit_query(self) -> str:
        return f"""WITH children AS (
                    SELECT id, parent_id, max_log_likelihood 
                    FROM fit WHERE {self.condition}
                   ), 
                   best AS (
                    SELECT parent_id, max(max_log_likelihood) AS max_log_likelihood 
                    FROM children 
                    GROUP BY parent_id
                   ) 
                   SELECT id 
                   FROM children, best 
                   WHERE children.parent_id = best.parent_id 
                   AND children.max_log_likelihood = best.max_log_likelihood;
               """



from typing import Optional

import numpy as np

from autofit.mapper.prior_model.prior_model import Model
from autofit.graphical.expectation_propagation import AbstractFactorOptimiser
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.prior_model import AbstractPriorModel
from autofit.non_linear.analysis import Analysis
from autofit.non_linear.paths.abstract import AbstractPaths
from .abstract import AbstractModelFactor


class FactorCallable:
    def __init__(
        self,
        prior_model: AbstractPriorModel,
        analysis: Analysis,
    ):
        self.prior_model = prior_model
        self.analysis = analysis

    def __call__(self, **kwargs: np.ndarray) -> float:
        """
        Creates an instance of the prior model and evaluates it, forming
        a factor.

        Parameters
        ----------
        kwargs
            Arguments with names that are unique for each prior.

        Returns
        -------
        Calculated likelihood
        """
        arguments = dict()
        for name_, array in kwargs.items():
            prior_id = int(name_.split("_")[1])
            prior = self.prior_model.prior_with_id(prior_id)
            arguments[prior] = array
        # noinspection PyTypeChecker
        instance = self.prior_model.instance_for_arguments(arguments)
        return self.analysis.log_likelihood_function(instance)

class AnalysisFactor(AbstractModelFactor):
    @property
    def prior_model(self):
        return self._prior_model

    def __init__(
        self,
        prior_model: AbstractPriorModel,
        analysis: Analysis,
        optimiser: Optional[AbstractFactorOptimiser] = None,
        name=None,
    ):
        """
        A factor in the graph that actually computes the likelihood of a model
        given values for each variable that model contains

        Parameters
        ----------
        prior_model
            A model with some dimensionality
        analysis
            A class that implements a function which evaluates how well an
            instance of the model fits some data
        optimiser
            A custom optimiser that will be used to fit this factor specifically
            instead of the default optimiser
        """
        self.label = prior_model.label
        self.analysis = analysis

        prior_variable_dict = {prior.name: prior for prior in prior_model.priors}

        super().__init__(
            prior_model=prior_model,
            factor=FactorCallable(
                prior_model=prior_model,
                analysis=analysis,
            ),
            optimiser=optimiser,
            prior_variable_dict=prior_variable_dict,
            name=name,
        )

    def tree_flatten(self):
        return (
            (self.prior_model,),
            (
                self.analysis,
                self.optimiser,
                self.name,
            ),
        )

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(
            children[0],
            analysis=aux_data[0],
            optimiser=aux_data[1],
            name=aux_data[2],
        )

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, item):
        try:
            return super().__getattr__(item)
        except AttributeError:
            return getattr(self.prior_model, item)

    def name_for_variable(self, variable):
        path = ".".join(self.prior_model.path_for_prior(variable))
        return f"{self.name}.{path}"

    def visualize(
        self, paths: AbstractPaths, instance: ModelInstance, during_analysis: bool
    ):
        """
        Visualise the instances provided using each factor.

        Instances in the ModelInstance must have the same order as the factors.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        instance
            A collection of instances, each corresponding to a factor
        during_analysis
            Is this visualisation during analysis?
        """
        self.analysis.visualize(
            paths=paths, instance=instance, during_analysis=during_analysis
        )

    def visualize_before_fit(
        self,
        paths: AbstractPaths,
        model: Model,
    ):
        """
        Visualise the model provided using each factor.

        Models in the ModelInstance must have the same order as the factors.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        model
            A collection of models, each corresponding to a factor
        """
        self.analysis.visualize_before_fit(paths=paths, model=model)

    def save_attributes(self, paths: AbstractPaths):
        """
        Save the attributes of the analysis object to a file.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        """
        self.analysis.save_attributes(paths=paths)

    def save_results(self, paths: AbstractPaths, result):
        """
        Save the results of the analysis to a file.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        result
            The result of the analysis
        """
        self.analysis.save_results(paths=paths, result=result)

    def log_likelihood_function(self, instance: ModelInstance) -> float:
        return self.analysis.log_likelihood_function(instance)
